import logging
import os.path as path
from typing import List, Tuple

from psychopy import core, visual, event

from bcipy.helpers.clock import Clock
from bcipy.helpers.task import get_key_press
from bcipy.helpers.symbols import SPACE_CHAR
from bcipy.display import (
    BCIPY_LOGO_PATH,
    Display,
    InformationProperties,
    PreviewInquiryProperties,
    StimuliProperties
)
from bcipy.display.components.task_bar import TaskBar
from bcipy.helpers.stimuli import resize_image
from bcipy.helpers.system_utils import get_screen_info
from bcipy.helpers.triggers import TriggerCallback, _calibration_trigger


class RSVPDisplay(Display):
    """RSVP Display Object for inquiry Presentation.

    Animates display objects common to any RSVP task.
    """

    def __init__(
            self,
            window: visual.Window,
            static_clock,
            experiment_clock: Clock,
            stimuli: StimuliProperties,
            task_bar: TaskBar,
            info: InformationProperties,
            preview_inquiry: PreviewInquiryProperties = None,
            trigger_type: str = 'image',
            space_char: str = SPACE_CHAR,
            full_screen: bool = False):
        """Initialize RSVP display parameters and objects.

        PARAMETERS:
        ----------
        # Experiment
        window(visual.Window): PsychoPy Window
        static_clock(core.MonotonicClock): Used to schedule static periods of display time
        experiment_clock(Clock): Clock used to timestamp display onsets

        # Stimuli
        stimuli(StimuliProperties): attributes used for inquiries

        # Task
        task_bar(TaskBar): used for task tracking. Ex. 1/100

        # Info
        info(InformationProperties): attributes to display informational stimuli alongside task and inquiry stimuli.

        # Preview Inquiry
        preview_inquiry(PreviewInquiryProperties) Optional: attributes to display a preview of upcoming stimuli defined
            via self.stimuli(StimuliProperties).

        trigger_type(str) default 'image': defines the calibration trigger type for the display at the beginning of any
            task. This will be used to reconcile timing differences between acquisition and the display.
        space_char(str) default SPACE_CHAR: defines the space character to use in the RSVP inquiry.
        full_screen(bool) default False: Whether or not the window is set to a full screen dimension. Used for
            scaling display items as needed.
        """
        self.window = window
        self.window_size = self.window.size  # [w, h]
        self.refresh_rate = window.getActualFrameRate()

        self.logger = logging.getLogger(__name__)

        # Stimuli parameters, these are set on display in order to allow
        #  easy updating after definition
        self.stimuli_inquiry = stimuli.stim_inquiry
        self.stimuli_colors = stimuli.stim_colors
        self.stimuli_timing = stimuli.stim_timing
        self.stimuli_font = stimuli.stim_font
        self.stimuli_height = stimuli.stim_height
        self.stimuli_pos = stimuli.stim_pos
        self.is_txt_stim = stimuli.is_txt_stim
        self.stim_length = stimuli.stim_length

        self.full_screen = full_screen
        self._preview_inquiry = preview_inquiry

        self.staticPeriod = static_clock

        # Trigger handling
        self.first_run = True
        self.first_stim_time = None
        self.trigger_type = trigger_type
        self.trigger_callback = TriggerCallback()
        self.experiment_clock = experiment_clock

        # Callback used on presentation of first stimulus.
        self.first_stim_callback = lambda _sti: None
        self.size_list_sti = []  # TODO force initial size definition
        self.space_char = space_char  # TODO remove and force task to define

        self.task_bar = task_bar

        # Create multiple text objects based on input
        self.info = info
        self.info_text = info.build_info_text(window)

        # Create initial stimuli object for updating
        self.sti = stimuli.build_init_stimuli(window)

    def draw_static(self):
        """Draw static elements in a stimulus."""
        if self.task_bar:
            self.task_bar.draw()
        for info in self.info_text:
            info.draw()

    def schedule_to(self,
                    stimuli: List[str] = None,
                    timing: List[float] = None,
                    colors: List[str] = None):
        """Schedule stimuli elements (works as a buffer).

        Args:
                stimuli(list[string]): list of stimuli text / name
                timing(list[float]): list of timings of stimuli
                colors(list[string]): list of colors
        """
        self.stimuli_inquiry = stimuli or []
        self.stimuli_timing = timing or []
        self.stimuli_colors = colors or []

    def do_inquiry(self, preview_calibration: bool = False) -> List[float]:
        """Do inquiry.

        Animates an inquiry of flashing letters to achieve RSVP.


        PARAMETERS:
        -----------
        preview_calibration(bool) default False: Whether or not to preview the upcoming inquiry stimuli. This feature
            is used to help the participant prepare for the upcoming inquiry after a prompt. It will present after
            the first stimulus of the inquiry (assumed to be a prompt). Not recommended for use outside of a
            calibration task.

        RETURNS:
        --------
        timing(list[float]): list of timings of stimuli presented in the inquiry
        """

        # init an array for timing information
        timing = []

        if self.first_run:
            self._trigger_pulse()

        # generate a inquiry (list of stimuli with meta information)
        inquiry = self._generate_inquiry()

        # do the inquiry
        for idx in range(len(inquiry)):

            # If this is the start of an inquiry and a callback registered for first_stim_callback evoke it
            if idx == 0 and callable(self.first_stim_callback):
                self.first_stim_callback(inquiry[idx]['sti'])

            # If previewing the inquiry, do so after the first stimulus
            if preview_calibration and idx == 1:
                time, _ = self.preview_inquiry()
                timing.extend(time)

            # Reset the timing clock to start presenting
            self.window.callOnFlip(
                self.trigger_callback.callback,
                self.experiment_clock,
                inquiry[idx]['sti_label'])

            # Draw stimulus for n frames
            inquiry[idx]['sti'].draw()
            self.draw_static()
            self.window.flip()
            core.wait(inquiry[idx]['time_to_present'])

            # append timing information
            timing.append(self.trigger_callback.timing)

            self.trigger_callback.reset()

        # draw in static and flip once more
        self.draw_static()
        self.window.flip()

        return timing

    def _trigger_pulse(self) -> None:
        """Trigger Pulse.

        This method uses a calibration trigger to determine any functional
            offsets needed for operation with this display. By setting the first_stim_time and searching for the
            same stimuli output to the marker stream, the offsets between these proceses can be reconciled at the
            beginning of an experiment. If drift is detected in your experiment, more frequent pulses and offset
            correction may be required.
        """
        calibration_time = _calibration_trigger(
            self.experiment_clock,
            trigger_type=self.trigger_type,
            display=self.window)

        # set the first stim time if not present and first_run to False
        if not self.first_stim_time:
            self.first_stim_time = calibration_time[-1]
            self.first_run = False

    def preview_inquiry(self) -> Tuple[List[float], bool]:
        """Preview Inquiry.

        Given an inquiry defined to be presented via do_inquiry(), present the full inquiry
            to the user and allow input on whether the intended letter is present or not before
            going through the rapid serial visual presention.

        Returns:
            - A tuple containing the timing information and a boolean describing whether to present
                the inquiry (True) or generate another (False).
        """
        # self._preview_inquiry defaults to None on __init__, assert it is defined correctly
        assert isinstance(self._preview_inquiry, PreviewInquiryProperties), (
            'PreviewInquiryProperties are not set on this RSVPDisplay. '
            'Add them as a preview_inquiry kwarg to use preview_inquiry().')
        # construct the timing to return and generate the content for preview
        timing = []
        if self.first_run:
            self._trigger_pulse()

        content = self._generate_inquiry_preview()

        # define the trigger callbacks. Here we use the trigger_callback to return immediately
        self.window.callOnFlip(
            self.trigger_callback.callback,
            self.experiment_clock,
            'inquiry_preview')

        # Draw and flip the screen.
        content.draw()
        self.draw_static()
        self.window.flip()
        timing.append(self.trigger_callback.timing)

        timer = core.CountdownTimer(self._preview_inquiry.preview_inquiry_length)
        response = False

        event.clearEvents(eventType='keyboard')
        while timer.getTime() > 0:
            # wait for a key press event
            response = get_key_press(
                key_list=[self._preview_inquiry.preview_inquiry_key_input],
                clock=self.experiment_clock,
            )

            # break if a response given unless this is preview only and wait the timer
            if response and not self._preview_inquiry.preview_only:
                break

        self.draw_static()
        self.window.flip()
        self.trigger_callback.reset()
        core.wait(self._preview_inquiry.preview_inquiry_isi)

        if self._preview_inquiry.preview_only:
            return timing, True

        # depending on whether or not press to accept, define what to return to the task
        if response and self._preview_inquiry.press_to_accept:
            timing.append(response)
            return timing, True
        elif response and not self._preview_inquiry.press_to_accept:
            timing.append(response)
            return timing, False
        elif not response and self._preview_inquiry.press_to_accept:
            return timing, False
        else:
            return timing, True

    def _generate_inquiry_preview(self) -> visual.TextBox2:
        """Generate Inquiry Preview.

        Using the self.stimuli_inquiry list, construct a preview box to display to the user. This method
            assumes the presence of a fixation (+).
        """
        text = ' '.join(self.stimuli_inquiry).split('+ ')[1]

        return self._create_stimulus(
            0.12,
            stimulus=text,
            units='height',
            stimuli_position=self.stimuli_pos,
            mode='textbox',
            align_text='left',
            wrap_width=0.025)

    def _generate_inquiry(self) -> list:
        """Generate inquiry.

        Generate stimuli for next RSVP inquiry. [A + A, C, Q, D]
        """
        stim_info = []
        for idx in range(len(self.stimuli_inquiry)):
            current_stim = {}

            current_stim['time_to_present'] = self.stimuli_timing[idx]

            # check if stimulus needs to use a non-default size
            if self.size_list_sti:
                this_stimuli_size = self.size_list_sti[idx]
            else:
                this_stimuli_size = self.stimuli_height

            # Set the Stimuli attrs
            if self.stimuli_inquiry[idx].endswith('.png'):
                current_stim['sti'] = self._create_stimulus(
                    mode='image',
                    height=this_stimuli_size,
                    stimulus=self.stimuli_inquiry[idx]
                )
                current_stim['sti'].size = resize_image(
                    current_stim['sti'].image, current_stim['sti'].win.size, this_stimuli_size)
                current_stim['sti_label'] = path.splitext(
                    path.basename(self.stimuli_inquiry[idx]))[0]
            else:
                # text stimulus
                current_stim['sti'] = self._create_stimulus(mode='text', height=this_stimuli_size)
                txt = self.stimuli_inquiry[idx]
                # customize presentation of space char.
                current_stim['sti'].text = txt if txt != SPACE_CHAR else self.space_char
                current_stim['sti'].color = self.stimuli_colors[idx]
                current_stim['sti_label'] = txt

                # test whether the word will be too big for the screen
                text_width = current_stim['sti'].boundingBox[0]
                if text_width > self.window.size[0]:
                    screen_info = get_screen_info()
                    monitor_width = screen_info.width
                    monitor_height = screen_info.height
                    text_height = current_stim['sti'].boundingBox[1]
                    # If we are in full-screen, text size in Psychopy norm units
                    # is monitor width/monitor height
                    if self.window.size[0] == monitor_width:
                        new_text_width = monitor_width / monitor_height
                    else:
                        # If not, text width is calculated relative to both
                        # monitor size and window size
                        new_text_width = (
                            self.window.size[1] / monitor_height) * (
                                monitor_width / monitor_height)
                    new_text_height = (text_height * new_text_width) / text_width
                    current_stim['sti'].height = new_text_height
            stim_info.append(current_stim)
        return stim_info

    def update_task_bar(self, text: str = None) -> None:
        """Update task state.

        Removes letters or appends to the right.
        Args:
                text(string): new text for task state
        """
        if self.task_bar:
            self.task_bar.update(text)

    def wait_screen(self, message: str, message_color: str) -> None:
        """Wait Screen.

        Args:
            message(string): message to be displayed while waiting
            message_color(string): color of the message to be displayed
        """

        # Construct the wait message
        wait_message = visual.TextStim(win=self.window,
                                       font=self.stimuli_font,
                                       text=message,
                                       height=.1,
                                       color=message_color,
                                       pos=(0, -.5),
                                       wrapWidth=2,
                                       colorSpace='rgb',
                                       opacity=1,
                                       depth=-6.0)

        # try adding the BciPy logo to the wait screen
        try:
            wait_logo = visual.ImageStim(
                self.window,
                image=BCIPY_LOGO_PATH,
                pos=(0, .25),
                mask=None,
                ori=0.0)
            wait_logo.size = resize_image(
                BCIPY_LOGO_PATH,
                self.window.size,
                1)
            wait_logo.draw()

        except Exception as e:
            self.logger.exception(f'Cannot load logo image from path=[{BCIPY_LOGO_PATH}]')
            raise e

        # Draw and flip the screen.
        wait_message.draw()
        self.window.flip()

    def _create_stimulus(
            self,
            height: int,
            mode: str = 'text',
            stimulus='+',
            color='white',
            stimuli_position=None,
            align_text='center',
            units=None,
            wrap_width=None,
            border=False):
        """Create Stimulus.

        Returns a TextStim, ImageStim or TextBox object.
        """
        if not stimuli_position:
            stimuli_position = self.stimuli_pos
        if mode == 'text':
            return visual.TextStim(
                win=self.window,
                color=color,
                height=height,
                text=stimulus,
                font=self.stimuli_font,
                pos=stimuli_position,
                wrapWidth=wrap_width,
                colorSpace='rgb',
                units=units,
                alignText=align_text,
                opacity=1,
                depth=-6.0)
        elif mode == 'image':
            return visual.ImageStim(
                win=self.window,
                image=stimulus,
                mask=None,
                units=units,
                pos=stimuli_position,
                size=(height, height),
                ori=0.0)
        elif mode == 'textbox':
            return visual.TextBox2(
                win=self.window,
                text=stimulus,
                color=color,
                colorSpace='rgb',
                borderWidth=2,
                borderColor='white',
                units=units,
                font=self.stimuli_font,
                letterHeight=height,
                size=[.5, .5],
                pos=stimuli_position,
                alignment=align_text,
            )



import unittest
import os
from unittest.mock import patch, mock_open

from collections import abc
import tempfile
import shutil
import json

from mockito import any, expect, unstub, when

from bcipy.config import (
    DEFAULT_ENCODING,
    DEFAULT_EXPERIMENT_PATH,
    DEFAULT_PARAMETERS_PATH,
    DEFAULT_FIELD_PATH,
    FIELD_FILENAME,
    EXPERIMENT_FILENAME
)
from bcipy.helpers.load import (
    extract_mode,
    load_json_parameters,
    load_experiments,
    load_experiment_fields,
    load_fields,
    load_users,
    copy_parameters)
from bcipy.helpers.parameters import Parameters
from bcipy.helpers.exceptions import BciPyCoreException, InvalidExperimentException


MOCK_EXPERIMENT = {
    "test": {
        "fields": [
            {
                "age": {
                    "required": "false",
                    "anonymize": "true"
                }
            }
        ],
        "summary": "test summary"
    }
}


class TestParameterLoad(unittest.TestCase):
    """This is Test Case for Loading BCI data."""

    def setUp(self):
        """set up the needed path for load functions."""

        self.parameters = DEFAULT_PARAMETERS_PATH
        self.temp_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.temp_dir)

    def test_load_json_parameters_returns_dict(self):
        """Test load parameters returns a Python dict."""

        # call the load parameters function
        parameters = load_json_parameters(self.parameters)

        # assert that load function turned json parameters into a dict-like obj
        self.assertEqual(type(parameters), Parameters)
        self.assertTrue(isinstance(parameters, abc.MutableMapping))

    def test_load_json_parameters_throws_error_on_wrong_path(self):
        """Test load parameters returns error on entering wrong path."""

        # call the load parameters function with incorrect path
        with self.assertRaises(Exception):
            load_json_parameters('/garbage/dir/wont/work')

    def test_copy_default_parameters(self):
        """Test that default parameters can be copied."""
        path = copy_parameters(destination=self.temp_dir)

        self.assertTrue(path != self.parameters)

        copy = load_json_parameters(path)
        self.assertTrue(type(copy), 'dict')

        parameters = load_json_parameters(self.parameters)
        self.assertEqual(copy, parameters)


class TestExperimentLoad(unittest.TestCase):

    def setUp(self):
        self.experiments_path = f'{DEFAULT_EXPERIMENT_PATH}/{EXPERIMENT_FILENAME}'
        when(json).load(any()).thenReturn()

    def tearDown(self):
        unstub()

    def test_load_experiments_calls_open_with_expected_default(self):
        with patch('builtins.open', mock_open(read_data='data')) as mock_file:
            load_experiments()
            mock_file.assert_called_with(self.experiments_path, 'r', encoding=DEFAULT_ENCODING)

    def test_load_experiments_throws_file_not_found_exception_with_invalid_path(self):
        with self.assertRaises(FileNotFoundError):
            load_experiments(path='')

    def test_load_experiments_calls_json_module_as_expected(self):
        with patch('builtins.open', mock_open(read_data='data')) as _:
            expect(json, times=1).loads(self.experiments_path)
            load_experiments()


class TestFieldLoad(unittest.TestCase):
    def setUp(self):
        self.fields_path = f'{DEFAULT_FIELD_PATH}/{FIELD_FILENAME}'
        when(json).load(any()).thenReturn()

    def tearDown(self):
        unstub()

    def test_load_fields_calls_open_with_expected_default(self):
        with patch('builtins.open', mock_open(read_data='data')) as mock_file:
            load_fields()
            mock_file.assert_called_with(self.fields_path, 'r', encoding=DEFAULT_ENCODING)

    def test_load_fields_throws_file_not_found_exception_with_invalid_path(self):
        with self.assertRaises(FileNotFoundError):
            load_fields(path='')

    def test_load_fields_calls_json_module_as_expected(self):
        with patch('builtins.open', mock_open(read_data='data')) as _:
            expect(json, times=1).loads(self.fields_path)
            load_fields()


class TestExperimentFieldLoad(unittest.TestCase):

    def setUp(self):
        self.experiment = MOCK_EXPERIMENT['test']

    def test_load_experiment_fields_returns_a_list(self):
        fields = load_experiment_fields(self.experiment)
        self.assertIsInstance(fields, list)

    def test_load_experiment_fields_raises_type_error_on_non_dict_experiment(self):
        invalid_experiment_type = ''
        with self.assertRaises(TypeError):
            load_experiment_fields(invalid_experiment_type)

    def test_load_experiment_fields_raises_invalid_experiment_on_incorrectly_formatted_experiment(self):
        # create an experiment with the wrong field key
        invalid_experiment_field_name = {
            'summary': 'blah',
            'field': []
        }
        with self.assertRaises(InvalidExperimentException):
            load_experiment_fields(invalid_experiment_field_name)


class TestUserLoad(unittest.TestCase):

    def setUp(self):
        # setup parameters to pass to load users method, it expects a key of data_save_loc only
        self.directory_name = 'test_data_load_user'
        self.data_save_loc = f'{self.directory_name}/'

    def test_user_load_with_no_directory_written(self):
        """Use defined data save location without writing anything"""
        response = load_users(self.data_save_loc)

        self.assertEqual(response, [])

    def test_user_load_with_valid_directory(self):
        user = 'user_001'
        file_path = f'{self.directory_name}/experiment/{user}'
        os.makedirs(file_path)

        response = load_users(self.data_save_loc)

        # There is only one user returned
        length_of_users = len(response)
        self.assertTrue(length_of_users == 1)

        # assert user returned is user defined above
        self.assertEqual(response[0], user)
        shutil.rmtree(self.data_save_loc)

    def test_user_load_with_invalid_directory(self):
        # create an invalid save structure and assert expected behavior.
        user = 'user_001'
        file_path = f'{self.directory_name}/experiment{user}'
        os.makedirs(file_path)

        response = load_users(self.data_save_loc)
        length_of_users = len(response)
        self.assertTrue(length_of_users == 0)
        shutil.rmtree(self.data_save_loc)


class TestExtractMode(unittest.TestCase):

    def test_extract_mode_calibration(self):
        data_save_path = 'data/default/user/user_RSVP_Calibration_Mon_01_Mar_2021_11hr19min49sec_-0800'
        expected_mode = 'calibration'
        response = extract_mode(data_save_path)
        self.assertEqual(expected_mode, response)

    def test_extract_mode_copy_phrase(self):
        data_save_path = 'data/default/user/user_RSVP_Copy_Phrase_Mon_01_Mar_2021_11hr19min49sec_-0800'
        expected_mode = 'copy_phrase'
        response = extract_mode(data_save_path)
        self.assertEqual(expected_mode, response)

    def test_extract_mode_without_mode_defined(self):
        invalid_data_save_dir = 'data/default/user/user_bad_dir'
        with self.assertRaises(BciPyCoreException):
            extract_mode(invalid_data_save_dir)


if __name__ == '__main__':
    unittest.main()



"""Defines helper methods and variables related to input symbols"""
import os
from string import ascii_uppercase

SPACE_CHAR = '_'
BACKSPACE_CHAR = '<'


def alphabet(parameters=None, include_path=True):
    """Alphabet.

    Function used to standardize the symbols we use as alphabet.

    Returns
    -------
        array of letters.
    """
    if parameters and not parameters['is_txt_stim']:
        # construct an array of paths to images
        path = parameters['path_to_presentation_images']
        stimulus_array = []
        for stimulus_filename in sorted(os.listdir(path)):
            # PLUS.png is reserved for the fixation symbol
            if stimulus_filename.endswith(
                    '.png') and not stimulus_filename.endswith('PLUS.png'):
                if include_path:
                    img = os.path.join(path, stimulus_filename)
                else:
                    img = os.path.splitext(stimulus_filename)[0]
                stimulus_array.append(img)
        return stimulus_array

    return list(ascii_uppercase) + [BACKSPACE_CHAR, SPACE_CHAR]


DEFAULT_SYMBOL_SET = alphabet()



"""Mock eye tracker that streams data using LabStreamingLayer"""

import logging
import math
import time

from numpy.random import uniform

from bcipy.acquisition.datastream.lsl_server import LslDataServer
from bcipy.acquisition.devices import DeviceSpec

log = logging.getLogger(__name__)


def eye_tracker_device() -> DeviceSpec:
    """Mock DeviceSpec for an eye tracker."""
    return DeviceSpec(name='EyeTracker',
                      channels=[
                          'leftEyeX', 'leftEyeY', 'rightEyeX', 'rightEyeY',
                          'leftPupilArea', 'rightPupilArea',
                          'pixelsPerDegreeX', 'pixelsPerDegreeY'
                      ],
                      sample_rate=500.0,
                      content_type='Gaze')


def eye_tracker_data_generator(display_x=1920, display_y=1080):
    """Generates sample eye tracker data.

    TODO: determine appropriate values for pixelsPerDegree fields.
    TODO: look info alternatives; maybe PyGaze.
    http://www.pygaze.org/about/
    """

    def area(diameter):
        return math.pi * (diameter / 2.0)**2

    while True:
        yield [
            float(int(uniform(0, display_x))),  # left eye x
            float(int(uniform(0, display_y))),  # left eye y
            float(int(uniform(0, display_x))),  # right eye x
            float(int(uniform(0, display_y))),  # right eye y
            area(uniform(2, 8)),  # left pupil area
            area(uniform(2, 8)),  # right pupil area
            uniform(1, 10),  # pixelsPerDegreeX
            uniform(1, 10)  # pixelsPerDegreeY
        ]


def eye_tracker_server() -> LslDataServer:
    """Create a demo lsl_server that serves eye tracking data."""

    return LslDataServer(device_spec=eye_tracker_device(),
                         generator=eye_tracker_data_generator())


def main():
    """Create an run an lsl_server"""
    try:
        server = eye_tracker_server()
        log.debug("New server created")
        server.start()
        log.debug("Server started")
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        print("Keyboard Interrupt")
        server.stop()


if __name__ == '__main__':
    main()



"""Tests for acquisition helper."""
import shutil
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

from bcipy.config import DEFAULT_PARAMETERS_PATH
from bcipy.helpers.acquisition import (init_device, init_eeg_acquisition,
                                       max_inquiry_duration, parse_stream_type,
                                       server_spec, stream_types)
from bcipy.helpers.load import load_json_parameters
from bcipy.helpers.save import init_save_data_structure


class TestAcquisition(unittest.TestCase):
    """Unit tests for acquisition helper"""

    def setUp(self):
        """set up the needed path for load functions."""
        self.parameters_used = DEFAULT_PARAMETERS_PATH
        self.parameters = load_json_parameters(self.parameters_used,
                                               value_cast=True)
        self.data_save_path = 'data/'
        self.user_information = 'test_user_001'
        self.task = 'RSVP Calibration'

        self.save = init_save_data_structure(self.data_save_path,
                                             self.user_information,
                                             self.parameters_used, self.task)

    def tearDown(self):
        """Override; teardown test"""
        shutil.rmtree(self.save)

    def test_init_acquisition(self):
        """Test init_eeg_acquisition with LSL client."""

        params = self.parameters
        params['acq_mode'] = 'EEG/DSI-24'

        client, servers = init_eeg_acquisition(params, self.save, server=True)

        client.stop_acquisition()
        client.cleanup()
        for server in servers:
            server.stop()

        self.assertEqual(1, len(servers))
        self.assertEqual(client.device_spec.name, 'DSI-24')

        self.assertTrue(Path(self.save, 'devices.json').is_file())

    def test_max_inquiry_duration(self):
        """Test the max inquiry duration function"""
        params = {
            'time_fixation': 0.5,
            'time_prompt': 1,
            'stim_length': 10,
            'time_flash': 0.25,
            'task_buffer_length': 0.75,
            'prestim_length': 0,
            'stim_jitter': 0
        }

        self.assertEqual(4.75, max_inquiry_duration(params))

    def test_parse_stream_types(self):
        """Test parsing the acq_mode parameter"""
        self.assertListEqual(['EEG', 'Gaze'], stream_types('EEG+Gaze'))
        self.assertListEqual(['EEG', 'Gaze'], stream_types(' EEG+Gaze'))
        self.assertListEqual(['EEG', 'Gaze'], stream_types('EEG + Gaze'))
        self.assertListEqual(['EEG', 'My stream'],
                             stream_types('EEG + My stream'),
                             "Should not validate the type itself.")

        self.assertListEqual(['EEG'], stream_types('EEG'))
        self.assertListEqual(['Gaze', 'EEG', 'EOG'],
                             stream_types('Gaze+EEG+EOG'))

        self.assertListEqual(['EEG', 'Gaze+'],
                             stream_types('EEG|Gaze+', delimiter='|'))

        self.assertListEqual(['EEG'], stream_types('EEG+EEG'))

    @patch('bcipy.helpers.acquisition.discover_device_spec')
    @patch('bcipy.helpers.acquisition.preconfigured_device')
    def test_init_device_using_discovery(self, preconfigured_device_mock,
                                         discover_spec_mock):
        """Test device initialization where spec is derived from the LSL
        stream."""
        device_name = 'TestDevice'
        device_type = 'EEG'

        device_mock = Mock()
        device_mock.name = device_name
        discover_spec_mock.return_value = device_mock
        preconfigured_device_mock.return_value = None

        device_spec = init_device(device_type)

        discover_spec_mock.assert_called_with(device_type)
        preconfigured_device_mock.assert_called_with(device_name, strict=False)
        self.assertEqual(device_spec, device_mock)

    @patch('bcipy.helpers.acquisition.discover_device_spec')
    @patch('bcipy.helpers.acquisition.preconfigured_device')
    def test_init_device_using_preconfigured(self, preconfigured_device_mock,
                                             discover_spec_mock):
        """Test device initialization where the discovered device_spec is
        overridden using a preconfigured device."""
        device_name = 'TestDevice'
        device_type = 'EEG'

        discovery_device_mock = Mock()
        discovery_device_mock.name = device_name

        device_mock = Mock()

        discover_spec_mock.return_value = discovery_device_mock
        preconfigured_device_mock.return_value = device_mock

        device_spec = init_device(device_type)

        discover_spec_mock.assert_called_with(device_type)
        preconfigured_device_mock.assert_called_with(device_name, strict=False)
        self.assertEqual(device_spec, device_mock)

    @patch('bcipy.helpers.acquisition.discover_device_spec')
    @patch('bcipy.helpers.acquisition.preconfigured_device')
    def test_init_device_with_named_device(self, preconfigured_device_mock,
                                           discover_spec_mock):
        """Test device initialization where preconfigured device name is
        is provided."""

        device_mock = Mock()
        preconfigured_device_mock.return_value = device_mock

        device_spec = init_device('EEG', 'DSI-24')

        discover_spec_mock.assert_not_called()
        preconfigured_device_mock.assert_called_with('DSI-24', strict=True)
        self.assertEqual(device_spec, device_mock)

    def test_parse_stream_type(self):
        """Test function to split the stream type into content_type, name"""
        self.assertEqual(('EEG', 'DSI-24'), parse_stream_type('EEG/DSI-24'))
        self.assertEqual(('Gaze', None), parse_stream_type('Gaze'))

    @patch('bcipy.helpers.acquisition.preconfigured_device')
    def test_server_spec_with_named_device(self, preconfigured_device_mock):
        """Test function to get the DeviceSpec for a mock server."""
        device_mock = Mock()
        preconfigured_device_mock.return_value = device_mock
        device_spec = server_spec(content_type='EEG', device_name='TestDevice')
        preconfigured_device_mock.assert_called_with('TestDevice', strict=True)
        self.assertEqual(device_spec, device_mock)

    @patch('bcipy.helpers.acquisition.with_content_type')
    def test_server_spec_without_named_device(self, with_content_type_mock):
        """Test function to get the DeviceSpec for a mock server."""
        device1 = Mock()
        device2 = Mock()
        with_content_type_mock.return_value = [device1, device2]

        device_spec = server_spec(content_type='EEG')
        with_content_type_mock.assert_called_with('EEG')
        self.assertEqual(device_spec, device1)


if __name__ == '__main__':
    unittest.main()



"""Helper functions for language model use."""
import math
import inspect
from typing import Dict, List, Tuple
import numpy as np
from bcipy.language.main import LanguageModel, ResponseType
from bcipy.helpers.symbols import alphabet
# pylint: disable=unused-import
# flake8: noqa

"""Only imported models will be included in language_models_by_name"""
from bcipy.language.model.uniform import UniformLanguageModel
from bcipy.language.model.causal import CausalLanguageModel
from bcipy.language.model.mixture import MixtureLanguageModel
from bcipy.language.model.kenlm import KenLMLanguageModel
from bcipy.language.model.unigram import UnigramLanguageModel

# flake8: noqa
from bcipy.helpers.exceptions import InvalidLanguageModelException


def language_models_by_name() -> Dict[str, LanguageModel]:
    """Returns available language models indexed by name."""
    return {lm.name(): lm for lm in LanguageModel.__subclasses__()}


def init_language_model(parameters: dict) -> LanguageModel:
    """
    Init Language Model configured in the parameters.

    Parameters
    ----------
        parameters : dict
            configuration details and path locations

    Returns
    -------
        instance of a LanguageModel
    """

    language_models = language_models_by_name()
    model = language_models[parameters.get("lang_model_type", "UNIFORM")]

    # introspect the model arguments to determine what parameters to pass.
    args = inspect.signature(model).parameters.keys()

    # select the relevant parameters into a dict.
    params = {key: parameters[key] for key in args & parameters.keys()}
    return model(response_type=ResponseType.SYMBOL,
                 symbol_set=alphabet(parameters),
                 **params)


def norm_domain(priors: List[Tuple[str, float]]) -> List[Tuple[str, float]]:
    """Convert a list of (symbol, likelihood) values from negative log
    likelihood to the probability domain (between 0 and 1)

    Parameters:
        priors - list of (symbol, likelihood) values.
            assumes that the units are in the negative log likelihood where
            the lowest value is the most likely.
    Returns:
        list of values in the probability domain (between 0 and 1),
            where the highest value is the most likely.
    """
    return [(sym, math.exp(-prob)) for sym, prob in priors]


def with_min_prob(symbol_probs: List[Tuple[str, float]],
                  sym_prob: Tuple[str, float]) -> List[Tuple[str, float]]:
    """Returns a new list of symbol-probability pairs where the provided
    symbol has a minimum probability given in the sym_prob.

    If the provided symbol is already in the list with a greater probability,
    the list of symbol_probs will be returned unmodified.

    If the new probability is added or modified, existing values are adjusted
    equally.

    Parameters:
    -----------
        symbol_probs - list of symbol, probability pairs
        sym_prob - (symbol, min_probability) defines the minimum probability
            for the given symbol in the returned list.

    Returns:
    -------
        list of (symbol, probability) pairs such that the sum of the
        probabilities is approx. 1.0.
    """
    new_sym, new_prob = sym_prob

    # Split out symbols and probabilities into separate lists, excluding the
    # symbol to be adjusted.
    symbols = []
    probs = []
    for sym, prob in symbol_probs:
        if sym != new_sym:
            symbols.append(sym)
            probs.append(prob)
        elif prob >= new_prob:
            # symbol prob in list is larger than minimum.
            return symbol_probs

    probabilities = np.array(probs)

    # Add new symbol and its probability
    all_probs = np.append(probabilities, new_prob / (1 - new_prob))
    all_symbols = symbols + [new_sym]

    normalized = all_probs / sum(all_probs)

    return list(zip(all_symbols, normalized))


def histogram(letter_prior: List[Tuple[str, float]]) -> str:
    """Given a list of letter, prob tuples, generate a histogram that can be
    output to the console.

    Parameters:
    -----------
        letter_prior - list of letter, probability pairs
    Returns:
    --------
        printable string which contains a histogram with the letter and probability as the label.
    """
    margin = "\t"
    star = '*'
    lines = []
    for letter, prob in sorted(letter_prior):
        units = int(round(prob * 100))
        lines.append(letter + ' (' + "%03.2f" % (prob) + ") :" + margin +
                     (units * star))
    return '\n'.join(lines)



"""Defines a RingBuffer with a fixed size; when full additional elements
overwrite the oldest items in the data structure.

Adapted from Python Cookbook by David Ascher, Alex Martelli
https://www.oreilly.com/library/view/python-cookbook/0596001673/ch05s19.html
"""


class RingBuffer:
    """Data structure with a fixed size; when full additional elements
    overwrite the oldest items in the data structure.

    Parameters
    ----------
        size_max - max size of the buffer
        pre_allocated - whether to create all values on initialization
        empty_value - if pre_allocated, empty_value is used to set the
            values with no data.
    """

    def __init__(self, size_max: int, pre_allocated: bool = False,
                 empty_value=None):
        assert size_max > 0
        self.empty_value = empty_value
        self.max = size_max
        self.data = [empty_value] * size_max if pre_allocated else []
        self.cur = 0
        self.full = False
        self.pre_allocated = pre_allocated

    def append(self, item):
        """Add an element to the buffer, overwriting if full."""
        if self.full or self.pre_allocated:
            # overwrite
            self.data[self.cur] = item
        else:
            self.data.append(item)
        if not self.full:
            self.full = self.cur == self.max - 1
        self.cur = (self.cur + 1) % self.max

    def get(self):
        """Return a list of elements from the oldest to the newest."""
        if self.full:
            return self.data[self.cur:] + self.data[:self.cur]
        return self.data

    def is_empty(self):
        return len(self.data) == 0 or self.data[0] == self.empty_value



import random
import shutil
import tempfile
import unittest

from pathlib import Path

import numpy as np

from bcipy.config import DEFAULT_PARAMETERS_PATH
from bcipy.acquisition.devices import DeviceSpec, register
from bcipy.helpers.copy_phrase_wrapper import CopyPhraseWrapper
from bcipy.helpers.load import load_json_parameters
from bcipy.helpers.symbols import alphabet
from bcipy.language.model.uniform import UniformLanguageModel
from bcipy.signal.model import PcaRdaKdeModel
from bcipy.task.data import EvidenceType


class TestCopyPhraseWrapper(unittest.TestCase):
    """Test CopyPhraseWrapper"""

    @classmethod
    def setUpClass(cls):
        np.random.seed(0)
        random.seed(0)
        cls.tmp_dir = Path(tempfile.mkdtemp())

        cls.device_spec = DeviceSpec(name="dummy_device", channels=["1", "2", "3"], sample_rate=10)
        register(cls.device_spec)

        cls.params_used = DEFAULT_PARAMETERS_PATH
        cls.params = load_json_parameters(cls.params_used, value_cast=True)

        # # Generate fake data and train a model
        # Specify data dimensions
        cls.num_trial = cls.params["stim_length"]
        cls.dim_x = 5
        cls.num_channel = cls.device_spec.channel_count
        cls.num_x_pos = 200
        cls.num_x_neg = 200

        # Generate Gaussian random data
        # TODO - deduplicate, add to test utils somewhere (data generation matching a device_spec)
        cls.pos_mean, cls.pos_std = 0, 0.5
        cls.neg_mean, cls.neg_std = 1, 0.5
        x_pos = cls.pos_mean + cls.pos_std * np.random.randn(cls.num_channel, cls.num_x_pos, cls.dim_x)
        x_neg = cls.neg_mean + cls.neg_std * np.random.randn(cls.num_channel, cls.num_x_neg, cls.dim_x)
        y_pos = np.ones(cls.num_x_pos)
        y_neg = np.zeros(cls.num_x_neg)

        # Stack and permute data
        x = np.concatenate([x_pos, x_neg], 1)
        y = np.concatenate([y_pos, y_neg], 0)
        permutation = np.random.permutation(cls.num_x_pos + cls.num_x_neg)
        x = x[:, permutation, :]
        y = y[permutation]

        cls.model = PcaRdaKdeModel(k_folds=10)
        cls.model.fit(x, y)

    @classmethod
    def tearDownClass(cls):
        shutil.rmtree(cls.tmp_dir)

    def test_valid_letters(self):
        alp = alphabet()
        cp = CopyPhraseWrapper(
            min_num_inq=1,
            max_num_inq=50,
            lmodel=None,
            device_spec=self.device_spec,
            signal_model=None,
            # fs=25,
            k=2,
            alp=alp,
            task_list=[("HELLO_WORLD", "HE")],
            is_txt_stim=True,
            evidence_names=[EvidenceType.LM, EvidenceType.ERP],
            stim_timing=[0.5, 0.25],
        )

        triggers = [
            ("+", 0.0),
            ("H", 0.5670222830376588),
            ("D", 0.8171830819919705),
            ("J", 1.0843321380089037),
            ("B", 1.3329724550130777),
            ("C", 1.5825864360085689),
            ("A", 1.833380013005808),
            ("F", 2.083211077027954),
            ("G", 2.333359022042714),
            ("I", 2.583265081048012),
            ("E", 2.833274284028448),
        ]
        target_info = [
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
        ]

        letters, timings, labels = cp.letter_info(triggers, target_info)
        expected_letters = ["H", "D", "J", "B", "C", "A", "F", "G", "I", "E"]
        expected_time = [
            0.5670222830376588,
            0.8171830819919705,
            1.0843321380089037,
            1.3329724550130777,
            1.5825864360085689,
            1.833380013005808,
            2.083211077027954,
            2.333359022042714,
            2.583265081048012,
            2.833274284028448,
        ]
        self.assertEqual(expected_letters, letters)
        self.assertEqual(expected_time, timings)
        self.assertEqual(len(letters), len(labels))

        triggers = [
            ("+", 0.1),
            ("H", 0.5670222830376588),
            ("D", 0.8171830819919705),
            ("J", 1.0843321380089037),
            ("B", 1.3329724550130777),
            ("C", 1.5825864360085689),
            ("A", 1.833380013005808),
            ("F", 2.083211077027954),
            ("G", 2.333359022042714),
            ("I", 2.583265081048012),
            ("E", 2.833274284028448),
        ]
        target_info = [
            "fixation",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
        ]
        letters, timings, labels = cp.letter_info(triggers, target_info)
        self.assertEqual(expected_letters, letters)
        self.assertEqual(expected_time, timings)
        self.assertEqual(["nontarget"] * (len(letters)), labels)

        # Test it throws an exception when letter is outside alphabet
        with self.assertRaises(Exception):
            cp.letter_info([("A", 0.0), ("*", 1.0)], ["nontarget", "nontarget"])

    def test_init_series_evaluate_inquiry(self):
        alp = alphabet()

        # Create fake data to provide as the user's response
        duration = int(self.num_trial * self.device_spec.sample_rate * self.params["trial_length"])
        response_eeg = self.pos_mean + self.pos_std * np.random.randn(self.num_channel, duration)

        # Note that frequencies for notch and bandpass filter below are chosen to be < 1/2 of sampling frequency
        # due to limitations of nyquist-shannon sampling theorem
        copy_phrase_task = CopyPhraseWrapper(
            min_num_inq=1,
            max_num_inq=50,
            lmodel=UniformLanguageModel(symbol_set=alp),
            device_spec=self.device_spec,
            signal_model=self.model,
            k=1,
            alp=alp,
            task_list=[("HELLO_WORLD", "HE")],
            is_txt_stim=True,
            evidence_names=[EvidenceType.LM, EvidenceType.ERP],
            stim_timing=[0.5, 0.25],
            notch_filter_frequency=4.0,
            filter_low=1.0,
            filter_high=4.0,
            stim_length=self.params["stim_length"],
        )

        is_accepted, sti = copy_phrase_task.initialize_series()
        self.assertFalse(is_accepted)

        triggers = [
            ("+", 0.0),
            ("H", 0.5670222830376588),
            ("D", 0.8171830819919705),
            ("J", 1.0843321380089037),
            ("B", 1.3329724550130777),
            ("C", 1.5825864360085689),
            ("A", 1.833380013005808),
            ("F", 2.083211077027954),
            ("G", 2.333359022042714),
            ("I", 2.583265081048012),
            ("E", 2.833274284028448),
        ]
        stimuli = [label for (label, _) in triggers]
        target_info = [
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
            "nontarget",
        ]

        new_series, inquiry = copy_phrase_task.evaluate_inquiry(
            response_eeg, triggers, target_info, self.params["trial_length"]
        )
        self.assertFalse(new_series)
        self.assertEqual(sorted(inquiry.stimuli[0]), sorted(stimuli))
        self.assertEqual(inquiry.durations[0], [self.params["time_fixation"]] +
                         [self.params["time_flash"]] * self.params["stim_length"])
        self.assertEqual(inquiry.colors[0], [self.params["fixation_color"]] +
                         [self.params["stim_color"]] * self.params["stim_length"])


if __name__ == '__main__':
    unittest.main()



import logging

from typing import List, Optional, Tuple
from pathlib import Path
from matplotlib.figure import Figure
import matplotlib.pyplot as plt

from bcipy.config import (TRIGGER_FILENAME, RAW_DATA_FILENAME,
                          DEFAULT_DEVICE_SPEC_FILENAME)
from bcipy.helpers.acquisition import analysis_channels
from bcipy.helpers.triggers import TriggerType, trigger_decoder
from bcipy.helpers.convert import convert_to_mne
from bcipy.helpers.load import choose_csv_file, load_raw_data
from bcipy.helpers.stimuli import mne_epochs
from bcipy.helpers.raw_data import RawData
from bcipy.signal.process import Composition, get_default_transform

import bcipy.acquisition.devices as devices
import mne
from mne import Epochs
from mne.io import read_raw_edf

log = logging.getLogger(__name__)


def visualize_erp(
        raw_data: RawData,
        channel_map: List[int],
        trigger_timing: List[float],
        trigger_labels: List[int],
        trial_length: float,
        transform: Optional[Composition] = None,
        plot_average: Optional[bool] = True,
        plot_topomaps: Optional[bool] = True,
        show: Optional[bool] = False,
        save_path: Optional[str] = None) -> List[Figure]:
    """ Visualize ERP.

    Generates a comparative ERP figure following a task execution. Given a set of trailed data,
    and labels describing two classes (Nontarget=0 and Target=1), they are plotted and may be saved
    or shown in a window.

    Returns a list of the figure handles created.

    PARAMETERS
    ----------
    raw_data(RawData): BciPy RawData
    channel_map(List[int]): Map of channels to remove (0) or keep (1)
    trigger_timing: list of trigger timing
    trigger_labels: list of trigger labels
    trial_length: length of trials described in trigger_timing / labels in seconds
    transform(Composition): Optional BciPy composition to apply to data before visualization
    plot_average: Optional[boolean]: whether or not to average over all channels. Default: True
    plot_topomap: Optional[boolean]: whether or not to plot all channels across target/nontarget. Default: True
    show: Optional[boolean]: whether or not to show the figures generated. Default: False
    save_path: optional path to a save location of the figure generated
    """
    mne_data = convert_to_mne(raw_data, channel_map=channel_map, transform=transform)
    epochs = mne_epochs(mne_data, trigger_timing, trial_length, trigger_labels)
    # *Note* We assume, as described above, two trigger classes are defined for use in trigger_labels
    # (Nontarget=0 and Target=1). This will map into two corresponding MNE epochs whose indexing starts at 1.
    # Therefore, epochs['1'] == Nontarget and epochs['2'] == Target.
    epochs = (epochs['1'], epochs['2'])
    figs = []
    if plot_average:
        figs.extend(visualize_evokeds(epochs, save_path=save_path, show=show))
    if plot_topomaps:
        figs.extend(visualize_joint_average(epochs, ['Non-Target', 'Target'], save_path=save_path, show=show))

    return figs


def plot_edf(edf_path: str, auto_scale: Optional[bool] = False):
    """Plot data from the raw edf file. Note: this works from an iPython
    session but seems to throw errors when provided in a script.

    Parameters
    ----------
        edf_path - full path to the generated edf file
        auto_scale - optional; if True will scale the EEG data; this is
            useful for fake (random) data but makes real data hard to read.
    """
    edf = read_raw_edf(edf_path, preload=True)
    if auto_scale:
        edf.plot(scalings='auto')
    else:
        edf.plot()


def visualize_csv_eeg_triggers(trigger_col: Optional[int] = None):
    """Visualize CSV EEG Triggers.

    This function is used to load in CSV data and visualize device generated triggers.

    Input:
        trigger_col(int)(optional): Column location of triggers in csv file.
            It defaults to the last column.

    Output:
        Figure of Triggers
    """
    # Load in CSV
    data = load_raw_data(choose_csv_file())
    raw_data = data.by_channel()

    # Pull out the triggers
    if not trigger_col:
        triggers = raw_data[-1]
    else:
        triggers = raw_data[trigger_col]

    # Plot the triggers
    plt.plot(triggers)

    # Add some titles and labels to the figure
    plt.title('Trigger Signal')
    plt.ylabel('Trigger Value')
    plt.xlabel('Samples')

    log.debug('Press Ctrl + C to exit!')
    # Show us the figure! Depending on your OS / IDE this may not close when
    #  The window is closed, see the message above
    plt.show()


def visualize_joint_average(
        epochs: Tuple[Epochs],
        labels: List[str],
        plot_joint_times: Optional[List[float]] = [-0.1, 0, 0.2, 0.3, 0.35, 0.4, 0.5],
        save_path: Optional[str] = None,
        show: Optional[bool] = False) -> List[Figure]:
    """Visualize Joint Average.

    Plots channel average for each condition and provides corresponding topomaps for times defined in plot_joint_times.

    See: https://mne.tools/dev/generated/mne.Evoked.html#mne.Evoked.plot_joint

    Note: assumes the number of Epochs (passed as a tuple) is equal to the number of labels provided
    """
    assert len(epochs) == len(labels), "The number of epochs must match labels in Visualize Joint Average"

    figs = []
    for i, label in enumerate(labels):
        avg = epochs[i].average()
        fig = avg.plot_joint(times=plot_joint_times, title=label, show=show)
        if save_path:
            fig.savefig(
                f'{save_path}/{label.lower()}_topomap.png'
            )
        figs.append(fig)
    return figs


def visualize_evokeds(epochs: Tuple[Epochs, Epochs], save_path: Optional[str]
                      = None, show: Optional[bool] = False) -> List[Figure]:
    """Visualize Evokeds.

    Using MNE Epochs, generate a compare evokeds plot using the mean and showing parametric confidence
    interval in shaded region.

    See: https://mne.tools/stable/generated/mne.viz.plot_compare_evokeds.html

    Note: Assumes first epoch is nontarget and second is target."""
    evokeds = dict(nontarget=list(epochs[0].iter_evoked()),
                   target=list(epochs[1].iter_evoked()))
    fig = mne.viz.plot_compare_evokeds(evokeds, combine='mean', show=show)
    if save_path:
        fig[0].savefig(f'{save_path}/average_erp.png')

    return fig


def visualize_session_data(session_path: str, parameters: dict, show=True) -> Figure:
    """Visualize Session Data.

    This method is used to load and visualize EEG data after a session.

    Currently, it plots target / nontarget ERPs. It filters out TriggerTypes event,
    fixation, preview, and bcipy internal triggers. It uses the filter parameters
    used during the session to filter the data before plotting.

    Input:
        session_path(str): Path to the session directory
        parameters(dict): Dictionary of parameters
        show(bool): Whether to show the figure

    Output:
        Figure of Session Data
    """
    # extract all relevant parameters
    poststim_length = parameters.get("trial_length")
    # get signal filtering information
    downsample_rate = parameters.get("down_sampling_rate")
    notch_filter = parameters.get("notch_filter_frequency")
    filter_high = parameters.get("filter_high")
    filter_low = parameters.get("filter_low")
    filter_order = parameters.get("filter_order")
    static_offset = parameters.get("static_trigger_offset")

    raw_data = load_raw_data(Path(session_path, f'{RAW_DATA_FILENAME}.csv'))
    channels = raw_data.channels
    sample_rate = raw_data.sample_rate

    devices.load(Path(session_path, DEFAULT_DEVICE_SPEC_FILENAME))
    device_spec = devices.preconfigured_device(raw_data.daq_type)

    # setup filtering
    default_transform = get_default_transform(
        sample_rate_hz=sample_rate,
        notch_freq_hz=notch_filter,
        bandpass_low=filter_low,
        bandpass_high=filter_high,
        bandpass_order=filter_order,
        downsample_factor=downsample_rate,
    )
    # Process triggers.txt files
    trigger_targetness, trigger_timing, _ = trigger_decoder(
        offset=static_offset,
        trigger_path=f"{session_path}/{TRIGGER_FILENAME}",
        exclusion=[TriggerType.PREVIEW, TriggerType.EVENT, TriggerType.FIXATION],
    )
    assert "nontarget" in trigger_targetness, "No nontarget triggers found."
    assert "target" in trigger_targetness, "No target triggers found."
    assert len(trigger_targetness) == len(trigger_timing), "Trigger targetness and timing must be the same length."

    labels = [0 if label == 'nontarget' else 1 for label in trigger_targetness]
    channel_map = analysis_channels(channels, device_spec)

    return visualize_erp(
        raw_data,
        channel_map,
        trigger_timing,
        labels,
        poststim_length,
        transform=default_transform,
        plot_average=True,
        plot_topomaps=True,
        save_path=session_path,
        show=show,
    )



# pylint: disable=E0611
import logging
import os
import sys
import re
from decimal import Decimal
from enum import Enum
from typing import Any, Callable, List, NamedTuple, Optional

from PyQt5.QtCore import pyqtSlot, Qt, QEvent, QTimer
from PyQt5.QtGui import QFont, QPixmap, QShowEvent
from PyQt5.QtWidgets import (QApplication, QCheckBox, QComboBox, QDoubleSpinBox,
                             QFileDialog, QHBoxLayout, QSpinBox, QLabel, QLineEdit,
                             QMessageBox, QPushButton, QScrollArea,
                             QVBoxLayout, QWidget)


def font(size: int = 14, font_family: str = 'Helvetica') -> QFont:
    """Create a Font object with the given parameters."""
    return QFont(font_family, size, QFont.Normal)


def invalid_length(min=1, max=25) -> bool:
    """Invalid Length.

    Returns a function, which when passed a string will assert whether a string meets min/max conditions.
    """
    return lambda string: len(string) < min or len(string) > max


def contains_whitespaces(string: str) -> bool:
    """Contains Whitespace.

    Checks for the presence of whitespace in a string.
    """
    return re.match(r'^(?=.*[\s])', string)


def contains_special_characters(string: str, regex: str = '[^0-9a-zA-Z_]+') -> bool:
    """Contains Special Characters.

    Checks for the presence of special chracters in a string. By default it will allow underscores.
    """
    disallowed_chars = re.compile(regex)
    return bool(disallowed_chars.search(string))


def static_text_control(parent,
                        label: str,
                        color: str = 'black',
                        size: int = 14,
                        font_family: str = 'Helvetica') -> QLabel:
    """Creates a static text control with the given font parameters. Useful for
    creating labels and help components."""
    static_text = QLabel(parent)
    static_text.setWordWrap(True)
    static_text.setText(label)
    static_text.setStyleSheet(f'color: {color};')
    static_text.setFont(font(size, font_family))
    return static_text


class AlertMessageType(Enum):
    """Alert Message Type.

    Custom enum used to abstract PyQT message types from downstream users.
    """
    WARN = QMessageBox.Warning
    QUESTION = QMessageBox.Question
    INFO = QMessageBox.Information
    CRIT = QMessageBox.Critical


class AlertResponse(Enum):
    """Alert Response.

    Custom enum used to abstract PyQT alert responses from downstream users.
    """
    OK = QMessageBox.Ok
    CANCEL = QMessageBox.Cancel
    YES = QMessageBox.Yes
    NO = QMessageBox.No


class PushButton(QPushButton):
    """PushButton.

    Custom Button to store unique identifiers which are required for coordinating
    events across multiple buttons."""
    id = None

    def get_id(self):
        if not self.id:
            raise Exception('No ID set on PushButton')

        return self.id


class ComboBox(QComboBox):
    """ComboBox with the same interface as a QLineEdit for getting and setting values.

    Parameters:
    -----------
      options - list of items to appear in the lookup
      selected_value - selected item; if this is not one of the options and new option will be
        created for this value.
    """

    def __init__(self, options: List[str], selected_value: str, **kwargs):
        super(ComboBox, self).__init__(**kwargs)
        self.options = options
        self.addItems(self.options)
        self.setText(selected_value)
        self.setEditable(True)

    def setText(self, value: str):
        """Sets the current index to the given value. If the value is not in the list of
        options it will be added."""
        if value not in self.options:
            self.options = [value] + self.options
            self.clear()
            self.addItems(self.options)
        self.setCurrentIndex(self.options.index(value))

    def text(self):
        """Gets the currentText."""
        return self.currentText()


class MessageBox(QMessageBox):
    """Message Box.

    A custom QMessageBox implementation to provide timeout functionality to QMessageBoxes.
    """

    def __init__(self, *args, **kwargs):
        QMessageBox.__init__(self, *args, **kwargs)

        self.timeout = 0
        self.current = 0

    def showEvent(self, event: QShowEvent) -> None:
        """showEvent.

        If a timeout greater than zero is defined, set a QTimer to call self.close after the defined timeout.
        """
        if self.timeout > 0:
            # timeout is in seconds (multiply by 1000 to get ms)
            QTimer().singleShot(self.timeout * 1000, self.close)
        super(MessageBox, self).showEvent(event)

    def setTimeout(self, timeout: float) -> None:
        """setTimeout.

        Setter for timeout variable. This keeps the pattern for setting on pyqt5 QMessageBoxes consistent. Timeout
            should be provided in seconds.
        """
        self.timeout = timeout


class AlertMessageResponse(Enum):
    """AlertMessageResponse.

    Defines the types of response buttons to give users receiving an AlertMessage.
    """

    OTE = "Okay to Exit"
    OCE = "Okay or Cancel Exit"


def alert_message(
        message: str,
        title: str = None,
        message_type: AlertMessageType = AlertMessageType.INFO,
        message_response: AlertMessageResponse = AlertMessageResponse.OTE,
        message_timeout: float = 0) -> MessageBox:
    """Constructs an Alert Message.

    Parameters
    ----------
        message - text to display
        title - optional title of the GUI window
        message_type - type of icon that is displayed
        message_response - response buttons available to the user
        message_time - optional timeout for displaying the alert
    """

    msg = MessageBox()
    if title:
        msg.setWindowTitle(title)
    msg.setText(message)
    msg.setIcon(message_type.value)
    msg.setTimeout(message_timeout)

    if message_response is AlertMessageResponse.OTE:
        msg.setStandardButtons(AlertResponse.OK.value)
    elif message_response is AlertMessageResponse.OCE:
        msg.setStandardButtons(AlertResponse.OK.value |
                               AlertResponse.CANCEL.value)
    return msg


class FormInput(QWidget):
    """A form element with a label, help tip, and input control. The default control is a text
    input. This object may be subclassed to specialize the input control or the arrangement of
    widgets. FormInputs have the ability to show and hide themselves.

    Parameters:
    ----------
        label - form label.
        value - initial value.
        help_tip - optional help text.
        options - optional list of recommended values used by some controls.
        help_size - font size for the help text.
        help_color - color of the help text.
        should_display - if False, the input is always hidden.
    """

    def __init__(self,
                 label: str,
                 value: str,
                 help_tip: str = None,
                 options: List[str] = None,
                 help_size: int = 12,
                 help_color: str = 'darkgray',
                 should_display: bool = True):
        super(FormInput, self).__init__()

        self.label = label
        self.help_tip = help_tip
        self.options = options
        self.should_display = should_display

        self.label_widget = self.init_label()
        self.help_tip_widget = self.init_help(help_size, help_color)
        self.control = self.init_control(value)
        self.control.installEventFilter(self)
        self.init_layout()

        if not should_display:
            self.hide()

    def eventFilter(self, source, event):
        """Event filter that suppresses the scroll wheel event."""
        if (event.type() == QEvent.Wheel and source is self.control):
            return True
        return False

    def init_label(self) -> QWidget:
        """Initialize the label widget."""
        return static_text_control(None, label=self.label)

    def init_help(self, font_size: int, color: str) -> QWidget:
        """Initialize the help text widget."""
        if self.help_tip and self.label != self.help_tip:
            return static_text_control(None,
                                       label=self.help_tip,
                                       size=font_size,
                                       color=color)
        return None

    def init_control(self, value) -> QWidget:
        """Initialize the form control widget.
        Parameter:
        ---------
            value - initial value
        """
        # Default is a text input
        return QLineEdit(value)

    def init_layout(self):
        """Initialize the layout by adding the label, help, and control widgets."""
        self.vbox = QVBoxLayout()
        if self.label_widget:
            self.vbox.addWidget(self.label_widget)
        if self.help_tip_widget:
            self.vbox.addWidget(self.help_tip_widget)
        self.vbox.addWidget(self.control)
        self.setLayout(self.vbox)

    def value(self) -> str:
        """Returns the value associated with the form input."""
        if self.control:
            return self.control.text()
        return None

    def cast_value(self) -> Any:
        """Returns the value associated with the form input, cast to the correct type.

        *If not defined by downstream classes, it will return the value.*
        """
        return self.value()

    def matches(self, term: str) -> bool:
        """Returns True if the input matches the given text, otherwise False."""
        text = term.lower()
        return (text in self.label.lower()) or (
            self.help_tip and
            text in self.help_tip.lower()) or text in self.value().lower()

    def show(self):
        """Show this widget, and all child widgets."""
        if self.should_display:
            for widget in self.widgets():
                if widget:
                    widget.setVisible(True)

    def hide(self):
        """Hide this widget, and all child widgets."""
        for widget in self.widgets():
            if widget:
                widget.setVisible(False)

    def widgets(self) -> List[QWidget]:
        """Returns a list of self and child widgets. List may contain None values."""
        return [self.label_widget, self.help_tip_widget, self.control, self]


class IntegerInput(FormInput):
    """FormInput to select a integer value using a spinbox for selection. Help text is not
    displayed for spinbox items.

    Parameters:
    ----------
        label - form label.
        value - initial value.
    """

    def __init__(self, **kwargs):
        super(IntegerInput, self).__init__(**kwargs)

    def init_control(self, value):
        """Override FormInput to create a spinbox."""
        spin_box = QSpinBox()
        spin_box.setMinimum(-100000)
        spin_box.setMaximum(100000)
        if value:
            spin_box.setValue(int(value))
        return spin_box

    def cast_value(self) -> str:
        """Override FormInput to return an integer value."""
        if self.control:
            return int(self.control.text())
        return None


class FloatInputProperties(NamedTuple):
    """Properties used when constructing a FloatInput component. Default values
    originate from the pyqt5 QDoubleSpinBox component defaults."""

    min: float = -sys.float_info.max  # 0.0 is the component default
    max: float = sys.float_info.max  # 99.99 is the component default
    decimals: int = 1
    step: float = 0.1


def float_input_properties(value: str) -> FloatInputProperties:
    """Given a string representation of a float value, determine suitable
    properties for the float component used to input or update this value.
    """
    # Determine from the component if there is a reasonable min or max constraint
    dec = Decimal(str(value))
    _sign, _digits, exponent = dec.as_tuple()
    if exponent > 0:
        return FloatInputProperties()
    return FloatInputProperties(decimals=abs(exponent), step=10**exponent)


class FloatInput(FormInput):
    """FormInput to select a float value using a spinbox for selection. Help text is not
    displayed for spinbox items.

    Parameters:
    ----------
        label - form label.
        value - initial value.
    """

    def __init__(self, **kwargs):
        super(FloatInput, self).__init__(**kwargs)

    def init_control(self, value):
        """Override FormInput to create a spinbox."""
        spin_box = QDoubleSpinBox()

        # Make a reasonable guess about precision and step size based on the initial value.
        props = float_input_properties(value)

        spin_box.setMinimum(props.min)
        spin_box.setMaximum(props.max)
        spin_box.setDecimals(props.decimals)
        spin_box.setSingleStep(props.step)
        spin_box.setValue(float(value))
        return spin_box

    def cast_value(self) -> float:
        """Override FormInput to return as a float."""
        if self.control:
            return float(self.control.text())
        return None


class BoolInput(FormInput):
    """FormInput to select a boolean value using a Checkbox. Help text is not
    displayed for checkbox items.

    Parameters:
    ----------
        label - form label.
        value - initial value.
    """

    def __init__(self, **kwargs):
        super(BoolInput, self).__init__(**kwargs)

    def init_label(self) -> QWidget:
        """Override. Checkboxes do not have a separate label."""
        return None

    def init_help(self, font_size: int, color: str) -> QWidget:
        """Override. Checkboxes do not display help."""
        return None

    def init_control(self, value):
        """Override to create a checkbox."""
        ctl = QCheckBox(self.label)
        ctl.setChecked(value == 'true')
        ctl.setFont(font())
        return ctl

    def value(self) -> str:
        return 'true' if self.control.isChecked() else 'false'


class SelectionInput(FormInput):
    """FormInput to select from a list of options. The options keyword
    parameter is required for this input.

    Parameters:
    ----------
        label - form label.
        value - initial value.
        help_tip - optional help text.
        options - list of recommended values.
        help_font_size - font size for the help text.
        help_color - color of the help text."""

    def __init__(self, **kwargs):
        assert isinstance(kwargs['options'], list), f"options are required for {kwargs['label']}"
        super(SelectionInput, self).__init__(**kwargs)

    def init_control(self, value) -> QWidget:
        """Override to create a Combobox."""
        return ComboBox(self.options, value)


class TextInput(FormInput):
    """FormInput for entering text.

    Parameters:
    ----------
        label - form label.
        value - initial value.
        help_tip - optional help text.
        help_font_size - font size for the help text.
        help_color - color of the help text."""

    def __init__(self, **kwargs):
        super(TextInput, self).__init__(**kwargs)


class FileInput(FormInput):
    """FormInput for selecting a file.

    Parameters:
    ----------
        label - form label.
        value - initial value.
        help_tip - optional help text.
        options - optional list of recommended values.
        help_font_size - font size for the help text.
        help_color - color of the help text.
    """

    def __init__(self, **kwargs):
        super(FileInput, self).__init__(**kwargs)

    def init_control(self, value) -> QWidget:
        """Override to create either a selection list or text field depending
        on whether there are recommended values."""
        if isinstance(self.options, list):
            return ComboBox(self.options, value)
        return QLineEdit(value)

    def init_button(self) -> QWidget:
        """Creates a Button to initiate the file/directory dialog."""
        btn = QPushButton('...')
        btn.setFixedWidth(40)
        btn.clicked.connect(self.update_path)
        return btn

    def prompt_path(self) -> str:
        """Prompts the user with a FileDialog. Returns the selected value or None."""

        name, _ = QFileDialog.getOpenFileName(caption='Select a file',
                                              filter='All Files (*)')
        return name

    def update_path(self) -> None:
        """Prompts the user for a value and updates the control's value if one was input."""
        name = self.prompt_path()
        if name:
            self.control.setText(name)

    def init_layout(self) -> None:
        """Overrides the layout to add the file dialog button."""
        self.button = self.init_button()
        self.vbox = QVBoxLayout()
        if self.label_widget:
            self.vbox.addWidget(self.label_widget)
        if self.help_tip_widget:
            self.vbox.addWidget(self.help_tip_widget)
        hbox = QHBoxLayout()
        hbox.addWidget(self.control)
        hbox.addWidget(self.button)
        self.vbox.addLayout(hbox)
        self.setLayout(self.vbox)

    def widgets(self) -> List[QWidget]:
        """Override to include button."""
        return super().widgets() + [self.button]


class DirectoryInput(FileInput):
    """Extends FileInput to prompt for a directory.

    Parameters:
    ----------
        label - form label.
        value - initial value.
        help_tip - optional help text.
        options - optional list of recommended values.
        help_font_size - font size for the help text.
        help_color - color of the help text.
    """

    def prompt_path(self):
        """Override to prompt for a directory."""
        return QFileDialog.getExistingDirectory(caption='Select a path')


class SearchInput(QWidget):
    """Search input widget. Consists of a text input field and a Clear button.
    Text changes to the input are passed to the on_search action.
    The cancel button clears the input.

    Parameters:
    -----------
        on_search - search function to call. Takes a single str parameter, the
            contents of the text box.
    """

    def __init__(self, on_search, font_size: int = 10):
        super(SearchInput, self).__init__()

        self.on_search = on_search

        self.hbox = QHBoxLayout()
        self.hbox.addWidget(static_text_control(None, label='Search: '))

        self.input = QLineEdit()
        self.input.setStyleSheet(
            f"font-size: {font_size}px; line-height: {font_size}px")
        self.input.textChanged.connect(self._search)
        self.hbox.addWidget(self.input)

        cancel_btn = QPushButton('Clear')
        cancel_btn.setStyleSheet(f"font-size: {font_size}px;")
        cancel_btn.clicked.connect(self._cancel)
        self.hbox.addWidget(cancel_btn)

        self.setLayout(self.hbox)

    def _search(self) -> None:
        self.on_search(self.input.text())

    def _cancel(self) -> None:
        self.input.clear()
        self.input.repaint()


class BCIGui(QWidget):
    """BCIGui.

    Primary GUI for downstream abstraction. Convenient handling of widgets and asset creation.
    """

    def __init__(self, title: str, width: int, height: int, background_color: str):
        super(BCIGui, self).__init__()
        logging.basicConfig(
            level=logging.DEBUG,
            format='%(name)s - %(levelname)s - %(message)s')
        self.logger = logging

        self.buttons = []
        self.input_text = []
        self.static_text = []
        self.images = []
        self.comboboxes = []
        self.widgets = []

        # set main window properties
        self.background_color = background_color
        self.window = QWidget()
        self.vbox = QVBoxLayout()
        self.setStyleSheet(f'background-color: {self.background_color};')

        self.timer = QTimer()

        self.title = title

        # determines height/width of window
        self.width = width
        self.height = height

        self.setWindowTitle(self.title)
        self.setFixedWidth(self.width)
        self.setFixedHeight(self.height)
        self.setLayout(self.vbox)

        self.create_main_window()

    def create_main_window(self) -> None:
        """Create Main Window.

        Construct the main window for display of assets.
        """
        self.window_layout = QHBoxLayout()
        self.window.setStyleSheet(f'background-color: {self.background_color};')
        self.window_layout.addWidget(self.window)
        self.vbox.addLayout(self.window_layout)

    def add_widget(self, widget: QWidget) -> None:
        """Add Widget.

        Add a Widget to the main GUI, alongside the main window.
        """
        widget_layout = QHBoxLayout()
        widget_layout.addWidget(widget)
        widget_layout.addStretch()
        self.vbox.addLayout(widget_layout)
        self.widgets.append(widget)

    def show_gui(self) -> None:
        """Show GUI.

        Build all registered assets and initialize the interface.
        """
        self.build_assets()
        self.init_ui()

    def build_buttons(self) -> None:
        """Build Buttons."""
        ...

    def build_images(self) -> None:
        """Build Images."""
        ...

    def build_text(self) -> None:
        """Build Text."""
        ...

    def build_inputs(self) -> None:
        """Build Inputs."""
        ...

    def build_assets(self) -> None:
        """Build Assets.

        Build add registered asset types.
        """
        self.build_text()
        self.build_inputs()
        self.build_buttons()
        self.build_images()

    def init_ui(self) -> None:
        """Initialize UI.

        This method sets up the grid and window. Finally, it calls the show method to make
            the window visible.
        """
        self.show()

    @pyqtSlot()
    def default_on_dropdown(self) -> None:
        """Default on dropdown.

        Default event to bind to comboboxes
        """
        pass

    @pyqtSlot()
    def default_button_clicked(self) -> None:
        """Default button clikced.

        The default action for buttons if none are registed.
        """
        sender = self.sender()
        self.logger.debug(sender.text() + ' was pressed')
        self.logger.debug(sender.get_id())

    def add_button(self,
                   message: str,
                   position: list,
                   size: list,
                   id: int = -1,
                   background_color: str = 'white',
                   text_color: str = 'default',
                   action: Optional[Callable] = None) -> PushButton:
        """Add Button."""
        btn = PushButton(message, self.window)
        btn.id = id
        btn.move(position[0], position[1])
        btn.resize(size[0], size[1])

        btn.setStyleSheet(f'background-color: {background_color}; color: {text_color};')

        if action:
            btn.clicked.connect(action)
        else:
            btn.clicked.connect(self.default_button_clicked)

        self.buttons.append(btn)
        return btn

    def add_combobox(self, position: list, size: list, items: list, background_color='default',
                     text_color='default',
                     action=None, editable=False) -> QComboBox:
        """Add combobox."""

        combobox = QComboBox(self.window)
        combobox.move(position[0], position[1])
        combobox.resize(size[0], size[1])

        if action:
            combobox.currentTextChanged.connect(action)
        else:
            combobox.currentTextChanged.connect(self.default_on_dropdown)

        combobox.setStyleSheet(f'background-color: {background_color}; color: {text_color};')

        if editable:
            combobox.setEditable(True)
        combobox.addItems(items)

        self.comboboxes.append(combobox)
        return combobox

    def add_image(self, path: str, position: list, size: int) -> QLabel:
        """Add Image."""
        if os.path.isfile(path):
            labelImage = QLabel(self.window)
            pixmap = QPixmap(path)
            # ensures the new label size will scale the image itself
            labelImage.setScaledContents(True)
            labelImage.setPixmap(pixmap)
            width = pixmap.width()
            height = pixmap.height()

            if width > height:
                new_width = size
                new_height = size * height / width
            else:
                new_height = size
                new_width = size * width / height

            labelImage.resize(new_width, new_height)
            labelImage.move(position[0], position[1])

            self.images.append(labelImage)
            return labelImage
        raise Exception('Invalid path to image provided')

    def add_static_textbox(self,
                           text: str,
                           position: list,
                           background_color: str = 'white',
                           text_color: str = 'default',
                           size: list = None,
                           font_family='Times',
                           font_size=12,
                           wrap_text=False) -> QLabel:
        """Add Static Text."""

        static_text = QLabel(self.window)
        static_text.setText(text)
        if wrap_text:
            static_text.setWordWrap(True)
        static_text.setStyleSheet(f'background-color: {background_color}; color: {text_color};')
        static_text.move(position[0], position[1])

        text_settings = QFont(font_family, font_size)
        static_text.setFont(text_settings)
        if size:
            static_text.resize(size[0], size[1])

        self.static_text.append(static_text)
        return static_text

    def add_text_input(self, position: list, size: list) -> QLineEdit:
        """Add Text Input."""
        textbox = QLineEdit(self.window)
        textbox.move(position[0], position[1])
        textbox.resize(size[0], size[1])

        self.input_text.append(textbox)
        return textbox

    def throw_alert_message(self,
                            title: str,
                            message: str,
                            message_type: AlertMessageType = AlertMessageType.INFO,
                            message_response: AlertMessageResponse = AlertMessageResponse.OTE,
                            message_timeout: float = 0) -> MessageBox:
        """Throw Alert Message."""
        msg = alert_message(message,
                            title=title,
                            message_type=message_type,
                            message_response=message_response,
                            message_timeout=message_timeout)
        return msg.exec_()

    def get_filename_dialog(self,
                            message: str = 'Open File',
                            file_type: str = 'All Files (*)',
                            location: str = "") -> str:
        """Get Filename Dialog."""
        file_name, _ = QFileDialog.getOpenFileName(self.window, message, location, file_type)
        return file_name


class ScrollableFrame(QWidget):
    """Scrollable Frame.

    A QWidget that constructs a scrollable frame and accepts another widget to display within the scrollable area.
    """

    def __init__(self, height: int, width: int, background_color: str = 'black', widget: QWidget = None):
        super().__init__()

        self.height = height
        self.width = width
        self.background_color = background_color
        self.setStyleSheet(f'background-color: {self.background_color}')

        # create a vertical box to wrap all other widgets / layouts
        self.vbox = QVBoxLayout()

        # create the scrollable are
        self.frame = QScrollArea()
        self.frame.setVerticalScrollBarPolicy(Qt.ScrollBarAlwaysOn)
        self.frame.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.frame.setWidgetResizable(True)
        self.frame.setFixedWidth(self.width)
        self.setFixedHeight(self.height)

        # if a widget is provided, add it to the scrollable frame
        if widget:
            self.widget = widget
            self.frame.setWidget(widget)

        # add the frame and set the layout
        self.vbox.addWidget(self.frame)
        self.setLayout(self.vbox)

        self.show()

    def refresh(self, new_widget: QWidget) -> None:
        """Refresh.

        Updates the scrollable area with a new or update widget.
        """
        self.vbox.removeWidget(self.frame)
        self.widget = new_widget
        self.frame.setWidget(self.widget)
        self.vbox.addWidget(self.frame)


class LineItems(QWidget):
    """Line Items.

    A QWidget that accepts a list of labels + button pairs. These items must have the following structure:
        {
            'item label': {
                'button1 label': {
                    'color': 'color of the button',
                    'textColor': 'color of the button text',
                    'action': action of button (a python method)
                },
                'button2 label': {
                    'color': 'color of the button',
                    'textColor': 'color of the button text',
                    'action': action of button (a python method)
                },
            }
        }
    """

    def __init__(self, items: List[dict], width: str):
        super().__init__()
        self.width = width
        self.items = items

        self.vbox = QVBoxLayout()
        self.setFixedWidth(self.width)

        # construct the line items as widgets added to the layout
        self.construct_line_items()

        # after constructing the line items, set the main layout and show the widget
        self.setLayout(self.vbox)
        self.show()

    def construct_line_items(self) -> None:
        """Construct Line Items.

        Loop through provided list and construct a layout with a label and buttons. Add to the main layout vbox.
        """
        for item in self.items:
            layout = QHBoxLayout()

            # item
            item_label = next(iter(item))
            label = QLabel(item_label)
            layout.addWidget(label)

            # buttons
            for key in item[item_label]:
                text_color = item[item_label][key]['textColor']
                color = item[item_label][key]['color']
                action = item[item_label][key]['action']
                button = PushButton(key)
                button.setStyleSheet(f'background-color: {color}; color: {text_color};')
                button.clicked.connect(action)

                if item[item_label][key].get('id'):
                    button.id = item[item_label][key]['id']
                layout.addWidget(button)

            self.vbox.addLayout(layout)


def app(args) -> QApplication:
    """Main app registry.

    Passes args from main and initializes the app
    """
    return QApplication(args)


def start_app() -> None:
    """Start BCIGui."""
    bcipy_gui = app(sys.argv)
    ex = BCIGui(title='BCI GUI', height=650, width=650, background_color='white')

    # ex.get_filename_dialog()
    # ex.add_button(message='Test Button', position=[200, 300], size=[100, 100], id=1)
    # ex.add_image(path='../static/images/gui/bci_cas_logo.png', position=[50, 50], size=200)
    # ex.add_static_textbox(
    #   text='Test static text',
    #   background_color='black',
    #   text_color='white',
    #   position=[100, 20],
    #   wrap_text=True)
    # ex.add_combobox(position=[100, 100], size=[100, 100], items=['first', 'second', 'third'], editable=True)
    # ex.add_text_input(position=[100, 100], size=[100, 100])
    ex.show_gui()
    ex.throw_alert_message(title='title', message='test', message_response=AlertMessageResponse.OCE, message_timeout=5)

    sys.exit(bcipy_gui.exec_())


if __name__ == '__main__':
    start_app()



from bcipy.acquisition.protocols.lsl.lsl_client import (LslAcquisitionClient,
                                                        discover_device_spec)
from bcipy.acquisition.datastream.lsl_server import LslDataServer, await_start
from bcipy.acquisition.multimodal import ClientManager

__all__ = [
    'LslAcquisitionClient', 'LslDataServer', 'await_start',
    'discover_device_spec', 'ClientManager'
]



import numpy as np
import logging
from typing import Dict, List
from copy import copy

log = logging.getLogger(__name__)


class DecisionCriteria:
    """Abstract class for Criteria which can be applied to evaluate a inquiry
    """

    def __init__(self, **kwargs):
        pass

    def reset(self):
        pass

    def decide(self, series: Dict):
        """
        Apply the given criteria.
        Parameters:
        -----------
            series - series data
                - target(str): target of the series
                - time_spent(ndarray[float]): |num_trials|x1
                      time spent on the inquiry
                - list_sti(list[list[str]]): presented symbols in each
                      inquiry
                - list_distribution(list[ndarray[float]]): list of |alp|x1
                        arrays with prob. dist. over alp

        """
        raise NotImplementedError()


class MinIterationsCriteria(DecisionCriteria):
    """ Returns true if the minimum number of iterations have not yet
        been reached. """

    def __init__(self, min_num_inq: int):
        """ Args:
                min_num_inq(int): minimum number of inquiry number before any
                 termination objective is allowed to be triggered """
        self.min_num_inq = min_num_inq

    def decide(self, series: Dict):
        # Note: we use 'list_sti' parameter since this is the number of
        # inquiries displayed. The length of 'list_distribution' is 1 greater
        # than this, since the language model distribution is added before
        # the first inquiry is displayed.
        current_inq = len(series['list_sti'])
        log.debug(
            f"Checking min iterations; current iteration is {current_inq}")
        return current_inq < self.min_num_inq


class DecreasedProbabilityCriteria(DecisionCriteria):
    """Returns true if the letter with the max probability decreased from the
        last inquiry."""

    def decide(self, series: Dict):
        if len(series['list_distribution']) < 2:
            return False
        prev_dist = series['list_distribution'][-2]
        cur_dist = series['list_distribution'][-1]
        return np.argmax(cur_dist) == np.argmax(
            prev_dist) and np.max(cur_dist) < np.max(prev_dist)


class MaxIterationsCriteria(DecisionCriteria):
    """Returns true if the max iterations have been reached."""

    def __init__(self, max_num_inq: int):
        """ Args:
                max_num_inq(int): maximum number of inquiries allowed before
                    mandatory termination """
        self.max_num_inq = max_num_inq

    def decide(self, series: Dict):
        # Note: len(series['list_sti']) != len(series['list_distribution'])
        # see MinIterationsCriteria comment
        current_inq = len(series['list_sti'])
        if current_inq >= self.max_num_inq:
            log.debug(
                "Committing to decision: max iterations have been reached.")
            return True
        return False


class ProbThresholdCriteria(DecisionCriteria):
    """Returns true if the commit threshold has been met."""

    def __init__(self, threshold: float):
        """ Args:
                threshold(float in [0,1]): A threshold on most likely
                    candidate posterior. If a candidate exceeds a posterior
                    the system terminates.
                 """
        assert 1 >= threshold >= 0, "stopping threshold should be in [0,1]"
        self.tau = threshold

    def decide(self, series: Dict):
        current_distribution = series['list_distribution'][-1]
        if np.max(current_distribution) > self.tau:
            log.debug("Committing to decision: posterior exceeded threshold.")
            return True
        return False


class MarginCriteria(DecisionCriteria):
    """ Stopping criteria based on the difference of two most likely candidates.
        This condition enforces the likelihood difference between two top
        candidates to be at least a value. E.g. in 4 category case with
        a margin 0.2, the edge cases [0.6,0.4,0.,0.] and [0.4,0.2,0.2,0.2]
        satisfy the condition.
    """

    def __init__(self, margin: float):
        """ Args:
                margin(float in [0,1]): Minimum distance required between
                    two most likely competing candidates to trigger termination.
                    """
        assert 1 >= margin >= 0, "difference margin should be in [0,1]"
        self.margin = margin

    def decide(self, series: Dict):
        # Get the current posterior probability values
        p = copy(series['list_distribution'][-1])
        # This criteria compares most likely candidates (best competitors)
        candidates = [p[idx] for idx in list(np.argsort(p)[-2:])]
        stopping_rule = np.abs(candidates[0] - candidates[1])
        d = stopping_rule > self.margin
        if d:
            log.debug("Committing to decision: margin is high enough.")

        return d


class MomentumCommitCriteria(DecisionCriteria):
    """ Stopping criteria based on Shannon entropy on the simplex
        Attr:
            lam(float): linear combination parameter between entropy and the
                speed term
            tau(float): decision threshold
            """

    def __init__(self, tau: float, lam: float):
        self.lam = lam
        self.tau = tau

    def reset(self):
        pass

    def decide(self, series):
        eps = np.power(.1, 6)

        prob_history = copy(series['list_distribution'])
        p = prob_history[-1]

        tmp_p = np.ones(len(p)) * (1 - self.tau) / (len(p) - 1)
        tmp_p[0] = self.tau

        tmp = -np.sum(p * np.log2(p + eps))
        tau_ = -np.sum(tmp_p * np.log2(tmp_p + eps))

        tmp_ = np.array(prob_history)
        mom_ = tmp_[1:] * (np.log(tmp_[1:] + eps) - np.log(tmp_[:-1] + eps))
        momentum = np.linalg.norm(mom_) / len(prob_history)

        if len(prob_history) > 1:
            stopping_rule = tmp - self.lam * momentum
        else:
            stopping_rule = tmp

        d = stopping_rule < tau_

        return d


class CriteriaEvaluator():
    """Evaluates whether an series should commit to a decision based on the
    provided criteria.

    Parameters:
    -----------
        continue_criteria: list of criteria; if any of these evaluate to true the
            decision maker continues.
        commit_criteria: list of criteria; if any of these return true and
            continue_criteria are all false, decision maker commits to a decision.
    """

    def __init__(self, continue_criteria: List[DecisionCriteria],
                 commit_criteria: List[DecisionCriteria]):
        self.continue_criteria = continue_criteria or []
        self.commit_criteria = commit_criteria or []

    @classmethod
    def default(cls, min_num_inq: int, max_num_inq: int, threshold: float):
        return cls(continue_criteria=[MinIterationsCriteria(min_num_inq)],
                   commit_criteria=[
                       MaxIterationsCriteria(max_num_inq),
                       ProbThresholdCriteria(threshold)
        ])

    def do_series(self):
        for el_ in self.continue_criteria:
            el_.reset()
        for el in self.commit_criteria:
            el.reset()

    def should_commit(self, series: Dict):
        """Evaluates the given series; returns true if stoppage criteria has
        been met, otherwise false.

        Parameters:
        -----------
            series - series data
                - target(str): target of the series
                - time_spent(ndarray[float]): |num_trials|x1
                      time spent on the inquiry
                - list_sti(list[list[str]]): presented symbols in each
                      inquiry
                - list_distribution(list[ndarray[float]]): list of |alp|x1
                        arrays with prob. dist. over alp
        """
        if any(
                criteria.decide(series)
                for criteria in self.continue_criteria):
            return False
        return any(
            criteria.decide(series) for criteria in self.commit_criteria)



"""DataAcquisitionClient for LabStreamingLayer data sources."""
import logging
from typing import List

from pylsl import StreamInfo, StreamInlet, local_clock, resolve_stream, resolve_byprop

from bcipy.acquisition.devices import DEFAULT_DEVICE_TYPE, DeviceSpec, IRREGULAR_RATE
from bcipy.acquisition.exceptions import InvalidClockError
from bcipy.acquisition.protocols.lsl.lsl_connector import (channel_names,
                                                           check_device)
from bcipy.acquisition.protocols.lsl.lsl_recorder import LslRecordingThread
from bcipy.acquisition.record import Record
from bcipy.helpers.clock import Clock
from bcipy.gui.viewer.ring_buffer import RingBuffer

log = logging.getLogger(__name__)


class LslAcquisitionClient:
    """Data Acquisition Client for devices streaming data using Lab Streaming
    Layer. Its primary use is dynamically querying streaming data in realtime,
    however, if the save_directory and filename parameters are provided it uses
    a LslRecordingThread to persist the data.

    Parameters
    ----------
        max_buffer_len: the maximum length of data to be queried. For continuously
            streaming data this is the number of seconds of data to retain. For
            irregular data, specify the number of samples. When using the RSVP
            paradigm, the max_buffer_len should be large enough to store data for
            the entire inquiry.
        device_spec: spec for the device from which to query data; if missing,
            this class will attempt to find the first EEG stream.
        save_directory: if present, persists the data to the given location.
        raw_data_file_name: if present, uses this name for the data file.
    """

    def __init__(self,
                 max_buffer_len: float = 1,
                 device_spec: DeviceSpec = None,
                 save_directory: str = None,
                 raw_data_file_name: str = None):
        super().__init__()
        self.device_spec = device_spec
        self.max_buffer_len = max_buffer_len

        self.experiment_clock = None

        self.inlet = None
        self._first_sample_time = None

        self.save_directory = save_directory
        self.raw_data_file_name = raw_data_file_name

        self.recorder = None
        self.buffer = None

    def start_acquisition(self) -> bool:
        """Connect to the datasource and start acquiring data.

        Returns
        -------
        bool : False if acquisition is already in progress, otherwise True.
        """
        if self.inlet:
            return False

        content_type = self.device_spec.content_type if self.device_spec else DEFAULT_DEVICE_TYPE
        streams = resolve_stream('type', content_type)
        if not streams:
            raise Exception(
                f'LSL Stream not found for content type {content_type}')
        stream_info = streams[0]

        self.inlet = StreamInlet(stream_info, max_buflen=self.max_buffer_len)

        if self.device_spec:
            check_device(self.device_spec, self.inlet.info())
        else:
            self.device_spec = device_from_metadata(self.inlet.info())

        if self.save_directory:
            self.recorder = LslRecordingThread(stream_info,
                                               self.save_directory,
                                               self.raw_data_file_name,
                                               self.device_spec)
            self.recorder.start()

        if self.max_buffer_len and self.max_buffer_len > 0:
            self.buffer = RingBuffer(size_max=self.max_samples)
        _, self._first_sample_time = self.inlet.pull_sample()
        return True

    @property
    def first_sample_time(self) -> float:
        """Timestamp returned by the first sample. If the data is being
        recorded this value reflects the timestamp of the first recorded sample"""
        if self.recorder:
            return self.recorder.first_sample_time
        return self._first_sample_time

    def stop_acquisition(self) -> None:
        """Disconnect from the data source."""
        log.debug("Stopping Acquisition...")
        if self.inlet:
            self.inlet.close_stream()
            self.inlet = None
        if self.recorder:
            self.recorder.stop()
            self.recorder.join()

        self.buffer = None

    def __enter__(self):
        """Context manager enter method that starts data acquisition."""
        self.start_acquisition()
        return self

    def __exit__(self, _exc_type, _exc_value, _traceback):
        """Context manager exit method to clean up resources."""
        self.stop_acquisition()

    def get_data(self,
                 start: float = None,
                 end: float = None,
                 limit: int = None) -> List[Record]:
        """Get data in time range.

        Parameters
        ----------
            start : starting timestamp (acquisition clock).
            end : end timestamp (in acquisition clock).
            limit: the max number of records that should be returned.

        Returns
        -------
            List of Records
        """
        log.debug(f"Getting data from: {start} to: {end} limit: {limit}")

        # Only data in the current buffer is available to query;
        # requests for data outside of this will fail. Buffer size is
        # set using the max_buffer_len parameter.
        data = self.get_latest_data()

        if not data:
            log.debug('No records available')
            return []

        log.debug((f'{len(data)} records available '
                   f'(From: {data[0].timestamp} To: {data[-1].timestamp})'))
        start = start or data[0].timestamp
        end = end or data[-1].timestamp
        limit = limit or -1
        assert start >= data[0].timestamp, (
            f'Start time of {start} is out of range: '
            f'({data[0].timestamp} to {data[-1].timestamp}).')

        data_slice = [
            record for record in data if start <= record.timestamp <= end
        ][0:limit]
        log.debug(f'{len(data_slice)} records returned')
        return data_slice

    @property
    def max_samples(self) -> int:
        """Maximum number of samples available at any given time."""
        if self.device_spec.sample_rate == IRREGULAR_RATE:
            return int(self.max_buffer_len)
        return int(self.max_buffer_len * self.device_spec.sample_rate)

    def get_latest_data(self) -> List[Record]:
        """Add all available samples in the inlet to the buffer.

        The number of items returned depends on the size of the configured
        max_buffer_len and the amount of data available in the inlet."""
        if not self.buffer:
            return []
        samples, timestamps = self.inlet.pull_chunk(
            max_samples=self.max_samples)

        for i, sample in enumerate(samples):
            self.buffer.append(
                Record(data=sample, timestamp=timestamps[i], rownum=None))

        return self.buffer.get()

    def convert_time(self, experiment_clock: Clock, timestamp: float) -> float:
        """
        Convert a timestamp from the experiment clock to the acquisition clock.
        Used for querying the acquisition data for a time slice.

        Parameters:
        ----------
        - experiment_clock : clock used to generate the timestamp
        - timestamp : timestamp from the experiment clock

        Returns:
        --------
            corresponding timestamp for the acquistion clock
        """

        # experiment_time = pylsl.local_clock() - offset
        return timestamp + self.clock_offset(experiment_clock)

    def get_data_seconds(self, seconds: int) -> List[Record]:
        """Returns the last n second of data"""
        assert seconds <= self.max_buffer_len, f"Seconds can't exceed {self.max_buffer_len}"

        sample_count = seconds * self.device_spec.sample_rate
        records = self.get_latest_data()

        start_index = 0 if len(
            records) > sample_count else len(records) - sample_count
        return records[start_index:]

    @property
    def is_calibrated(self):
        """Returns boolean indicating whether or not acquisition has been
        calibrated (an offset calculated based on a trigger)."""
        return True

    @is_calibrated.setter
    def is_calibrated(self, bool_val):
        """Setter for the is_calibrated property that allows the user to
        override the calculated value and use a 0 offset.

        Parameters
        ----------
        - bool_val : boolean
        if True, uses a 0 offset; if False forces the calculation.
        """

    def clock_offset(self, experiment_clock: Clock = None) -> float:
        """
        Offset in seconds from the experiment clock to the acquisition local clock.

        The experiment clock should be monotonic from experiment start time.
        The acquisition clock (pylsl.local_clock()) is monotonic from local
        machine start time (or since 1970-01-01 00:00). Therefore the acquisition
        clock should always be greater than experiment clock. An exception is
        raised if this doesn't hold.

        See https://labstreaminglayer.readthedocs.io/info/faqs.html#lsl-local-clock
        """
        clock = experiment_clock or self.experiment_clock
        assert clock, "An experiment clock must be provided"

        diff = local_clock() - clock.getTime()
        if diff < 0:
            raise InvalidClockError(
                "The acquisition clock should always be greater than experiment clock"
            )
        return diff

    def event_offset(self, event_clock: Clock, event_time: float) -> float:
        """Compute number of seconds that recording started prior to the given
        event.

        Parameters
        ----------
        - event_clock : monotonic clock used to record the event time.
        - event_time : timestamp of the event of interest.

        Returns
        -------
        Seconds between acquisition start and the event.
        """
        if self.first_sample_time:
            lsl_event_time = self.convert_time(event_clock, event_time)
            return lsl_event_time - self.first_sample_time
        return 0.0

    def offset(self, first_stim_time: float) -> float:
        """Offset in seconds from the start of acquisition to the given stim
        time.

        Parameters
        ----------
        - first_stim_time : LSL local clock timestamp of the first stimulus.

        Returns
        -------
        The number of seconds between acquisition start and the calibration
        event, or 0.0 .
        """

        if not first_stim_time:
            return 0.0
        assert self.first_sample_time, "Acquisition was not started."
        offset_from_stim = first_stim_time - self.first_sample_time
        log.debug(f"Acquisition offset: {offset_from_stim}")
        return offset_from_stim

    def cleanup(self):
        """Perform any necessary cleanup."""


def discover_device_spec(content_type: str) -> DeviceSpec:
    """Finds the first LSL stream with the given content type and creates a
    device spec from the stream's metadata."""
    log.info(f"Waiting for {content_type} data to be streamed over LSL.")
    streams = resolve_byprop('type', content_type, timeout=5.0)
    if not streams:
        raise Exception(
            f'LSL Stream not found for content type {content_type}')
    stream_info = streams[0]
    inlet = StreamInlet(stream_info)
    spec = device_from_metadata(inlet.info())
    inlet.close_stream()
    return spec


def device_from_metadata(metadata: StreamInfo) -> DeviceSpec:
    """Create a device_spec from the data stream metadata."""
    return DeviceSpec(name=metadata.name(),
                      channels=channel_names(metadata),
                      sample_rate=metadata.nominal_srate(),
                      content_type=metadata.type())



from bcipy.signal.process.decomposition.cwt import continuous_wavelet_transform
from bcipy.signal.process.decomposition.psd import power_spectral_density

__all__ = [
    "continuous_wavelet_transform",
    "power_spectral_density"
]



"""Module for functionality related to session-related data."""
from collections import Counter
from enum import Enum
from typing import Any, Dict, List, Optional

EVIDENCE_SUFFIX = "_evidence"


def rounded(values: List[float], precision: int) -> List[float]:
    """Round the list of values to the given precision.

    Parameters
    ----------
        values - values to round
    """
    return [round(value, precision) for value in values]


class EvidenceType(Enum):
    """Enum of the supported evidence types used in the various spelling tasks."""
    LM = 'LM'  # Language Model
    ERP = 'ERP'  # Event-Related Potential using EEG signals
    BTN = 'BTN'  # Button

    @classmethod
    def list(cls) -> List[str]:
        """List of evidence types"""
        return [ev_type.name for ev_type in cls]

    @classmethod
    def deserialized(cls, serialized_name: str) -> 'EvidenceType':
        """Deserialized name of the given evidence type.
        Parameters:
            evidence_name - ex. 'lm_evidence'
        Returns:
            deserialized value: ex. EvidenceType.LM
        """
        if serialized_name == 'eeg_evidence':
            return EvidenceType.ERP
        return cls(serialized_name[:-len(EVIDENCE_SUFFIX)].upper())

    def __str__(self) -> str:
        return self.name

    @property
    def serialized(self) -> str:
        """Name used when serialized to a json file."""
        if self == EvidenceType.ERP:
            return 'eeg_evidence'
        return f'{self.name.lower()}{EVIDENCE_SUFFIX}'


class Inquiry:
    """Represents a sequence of stimuli.

    Parameters:
    ----------
        stimuli - list of stimuli presented (letters, icons, etc)
        timing - duration in seconds for each stimulus
        target_info - targetness ('nontarget', 'target', etc) for each stimulus
        target_letter - current letter that the user is attempting to spell
        current_text - letters spelled so far
        target_text - word or words the user is attempting to spell
        next_display_state - text to be displayed after evaluating the current evidence
        lm_evidence - language model evidence for each stimulus
        eeg_evidence - eeg evidence for each stimulus
        likelihood - combined likelihood for each stimulus
    """

    def __init__(self,
                 stimuli: List[str],
                 timing: List[float],
                 triggers: List[List],
                 target_info: List[str],
                 target_letter: str = None,
                 current_text: str = None,
                 target_text: str = None,
                 selection: str = None,
                 next_display_state: str = None,
                 likelihood: List[float] = None):
        super().__init__()
        self.stimuli = stimuli
        self.timing = timing
        self.triggers = triggers
        self.target_info = target_info
        self.target_letter = target_letter
        self.current_text = current_text
        self.target_text = target_text
        self.selection = selection
        self.next_display_state = next_display_state

        self.evidences: Dict[EvidenceType, List[float]] = {}
        self.likelihood = likelihood or []
        # Precision used for serialization of evidence values.
        self.precision = None

    @property
    def lm_evidence(self):
        """Language model evidence"""
        return self.evidences.get(EvidenceType.LM, [])

    @property
    def eeg_evidence(self):
        """EEG evidence"""
        return self.evidences.get(EvidenceType.ERP, [])

    @property
    def decision_made(self) -> bool:
        """Returns true if the result of the inquiry was a decision."""
        return self.current_text != self.next_display_state

    @property
    def is_correct_decision(self) -> bool:
        """Indicates whether the current selection was the target"""
        return self.selection and (self.selection == self.target_letter)

    @classmethod
    def from_dict(cls, data: dict):
        """Deserializes from a dict

        Parameters:
        ----------
            data - a dict in the format of the data output by the as_dict
                method.
        """
        # partition into evidence data and other data.

        evidences = {
            EvidenceType.deserialized(name): value
            for name, value in data.items() if name.endswith(EVIDENCE_SUFFIX)
        }

        non_evidence_data = {
            name: value
            for name, value in data.items()
            if not name.endswith(EVIDENCE_SUFFIX)
        }
        inquiry = cls(**non_evidence_data)
        if len(data['stimuli']) == 1 and isinstance(data['stimuli'][0], list):
            # flatten
            inquiry.stimuli = data['stimuli'][0]
        inquiry.evidences = evidences
        return inquiry

    def as_dict(self) -> Dict:
        """Dict representation"""
        data = {
            'stimuli': self.stimuli,
            'timing': self.timing,
            'triggers': self.triggers,
            'target_info': self.target_info,
            'target_letter': self.target_letter,
            'current_text': self.current_text,
            'target_text': self.target_text,
            'selection': self.selection,
            'next_display_state': self.next_display_state
        }

        for evidence_type, evidence in self.evidences.items():
            data[evidence_type.serialized] = self.format(evidence)

        if self.likelihood:
            data['likelihood'] = self.format(self.likelihood)
        return data

    def stim_evidence(self,
                      symbol_set: List[str],
                      n_most_likely: int = 5) -> Dict[str, Any]:
        """Returns a dict of stim sequence data useful for debugging. Evidences
        are paired with the appropriate symbol for easier visual
        scanning. Also, an additional attribute is provided to display the
        top n most likely symbols based on the current evidence.

        Parameters:
        -----------
            symbol_set - list of stim in the same order as the evidences.
            n_most_likely - number of most likely elements to include
        """
        likelihood = dict(zip(symbol_set, self.format(self.likelihood)))
        data: Dict[str, Any] = {
            'stimuli': self.stimuli,
        }
        for evidence_type, evidence in self.evidences.items():
            data[evidence_type.serialized] = dict(
                zip(symbol_set, self.format(evidence)))
        data['likelihood'] = likelihood
        data['most_likely'] = dict(
            Counter(likelihood).most_common(n_most_likely))
        return data

    def format(self, evidence: List[float]) -> List[float]:
        """Format the evidence for output.

        Parameters
        ----------
            evidence - list of evidence values
        """
        if self.precision:
            return rounded(evidence, self.precision)
        return evidence


class Session:
    """Represents a data collection session. Not all tasks record session data."""

    def __init__(self,
                 save_location: str,
                 task: str = 'Copy Phrase',
                 mode: str = 'RSVP',
                 symbol_set: List[str] = None,
                 decision_threshold: float = None):
        super().__init__()
        self.save_location = save_location
        self.task = task
        self.mode = mode
        self.series: List[List[Inquiry]] = [[]]
        self.total_time_spent = 0
        self.time_spent_precision = 2
        self.task_summary = {}
        self.symbol_set = symbol_set
        self.decision_threshold = decision_threshold

    @property
    def total_number_series(self) -> int:
        """Total number of series that contain sequences."""
        return len([lst for lst in self.series if lst])

    @property
    def total_number_decisions(self) -> int:
        """Total number of series that ended in a decision."""
        # An alternate implementation would be to count the inquiries with
        # decision_made property of true.
        return len(self.series) - 1

    @property
    def total_inquiries(self) -> int:
        """Total number of inquiries presented."""
        return sum([len(lst) for lst in self.series])

    @property
    def inquiries_per_selection(self) -> Optional[float]:
        """Inquiries per selection"""
        selections = self.total_number_decisions
        if selections == 0:
            return None
        return self.total_inquiries / selections

    @property
    def all_inquiries(self) -> List[Inquiry]:
        """List of all Inquiries for the whole session"""
        return [inq for inquiries in self.series for inq in inquiries if inquiries]

    def has_evidence(self) -> bool:
        """Tests whether any inquiries have evidence."""
        return any(inq.evidences for inq in self.all_inquiries)

    def add_series(self):
        """Add another series unless the last one is empty"""
        if self.last_series():
            self.series.append([])

    def add_sequence(self, inquiry: Inquiry, new_series: bool = False):
        """Append sequence information

        Parameters:
        -----------
            inquiry - data to append
            new_series - a True value indicates that this is the first stim of
                a new series.
        """
        if new_series:
            self.add_series()
        self.last_series().append(inquiry)

    def last_series(self) -> List[Inquiry]:
        """Returns the last series"""
        return self.series[-1]

    def last_inquiry(self) -> Optional[Inquiry]:
        """Returns the last inquiry of the last series."""
        series = self.last_series()
        if series:
            return series[-1]
        return None

    def latest_series_is_empty(self) -> bool:
        """Whether the latest series has had any inquiries added to it."""
        return len(self.last_series()) == 0

    def as_dict(self,
                evidence_only: bool = False) -> Dict[str, Any]:
        """Dict representation"""
        series_dict: Dict[str, Any] = {}
        for i, series in enumerate(self.series):
            if series:
                series_counter = str(i + 1)
                series_dict[series_counter] = {}
                for series_index, stim in enumerate(series):
                    if evidence_only and self.symbol_set:
                        stim_dict = stim.stim_evidence(self.symbol_set)
                    else:
                        stim_dict = stim.as_dict()
                    series_dict[series_counter][str(series_index)] = stim_dict

        info = {
            'session': self.save_location,
            'task': self.task,
            'mode': self.mode,
            'symbol_set': self.symbol_set,
            'decision_threshold': self.decision_threshold,
            'series': series_dict,
            'total_time_spent': round(self.total_time_spent,
                                      self.time_spent_precision),
            'total_minutes': round(self.total_time_spent / 60, self.time_spent_precision),
            'total_number_series': self.total_number_series,
            'total_inquiries': self.total_inquiries,
            'total_selections': self.total_number_decisions,
            'inquiries_per_selection': self.inquiries_per_selection
        }

        if self.task_summary:
            info['task_summary'] = self.task_summary
        return info

    @classmethod
    def from_dict(cls, data: dict):
        """Deserialize from a dict.

        Parameters:
        ----------
            data - a dict in the format of the data output by the as_dict
                method.
        """
        session = cls(save_location=data['session'],
                      task=data['task'],
                      mode=data['mode'],
                      symbol_set=data['symbol_set'],
                      decision_threshold=data['decision_threshold'])
        session.total_time_spent = data['total_time_spent']

        if data['series']:
            session.series.clear()

            for series_key in sorted(data['series'].keys(), key=int):
                session.series.append([])

                for inquiry_key in sorted(data['series'][series_key], key=int):
                    inquiry_dict = data['series'][series_key][inquiry_key]
                    session.add_sequence(Inquiry.from_dict(inquiry_dict))

        return session



"""Streams data from pylsl and puts it into a Queue."""
import pylsl
from bcipy.acquisition.protocols.lsl.lsl_client import device_from_metadata
from bcipy.gui.viewer.data_source.data_source import DataSource


class LslDataSource(DataSource):
    """DataSource that provides data from an underlying pylsl StreamInlet.

    Parameters
    ----------
    - stream_type: StreamInlet stream type; default is 'EEG'
    """

    def __init__(self,
                 stream_type: str = 'EEG',
                 max_timeout_seconds: float = 1.0):
        super(LslDataSource, self).__init__()

        print('Waiting for LSL EEG data stream...')
        self.stream_type = stream_type
        self.max_timeout_seconds = max_timeout_seconds
        streams = pylsl.resolve_stream('type', self.stream_type)
        self.inlet = pylsl.StreamInlet(streams[0])
        self.device_spec = device_from_metadata(self.inlet.info())

    def next(self):
        """Provide the next record."""
        sample, _ts = self.inlet.pull_sample(timeout=0.0)
        return sample

    def next_n(self, n: int, fast_forward=False):
        """Provides the next n records as a list

        Parameters:
        -----------
            n - number of records to retrieve from LSL
            fast_forward - if true, fast forwards to the latest data. This
                flag should be used when first connecting to LSL or resuming
                from a paused state. Otherwise the data returned will be the
                next n records from the last point consumed.
        """
        # Read data from the inlet. May blocks GUI interaction if n samples
        # are not yet available.
        samples, _ts = self.inlet.pull_chunk(timeout=self.max_timeout_seconds,
                                             max_samples=n)

        if fast_forward:
            tmp = samples
            print('Fast forwarding:')
            chomped_count = 0
            while len(tmp) == n:
                samples = tmp
                # A timeout of 0.0 does not block GUI interaction and only
                # gets samples immediately available.
                tmp, _ts = self.inlet.pull_chunk(timeout=0.0, max_samples=n)
                chomped_count += len(tmp)
            print(f'Chomped {chomped_count} records.')
            samples = samples[len(tmp):] + tmp

        if len(samples) < n and not fast_forward:
            # Fast forwarding generally occurs when the stream is first started
            # or when resuming after a pause, so there may not be a full n
            # samples available. However, if we are streaming normally and
            # less than n samples are returned, we assume the stream has
            # terminated.
            raise StopIteration
        return samples



"""Defines the Copy Phrase Task which uses a Matrix display"""

from bcipy.task.paradigm.rsvp.copy_phrase import RSVPCopyPhraseTask
from bcipy.display import Display, InformationProperties, StimuliProperties
from bcipy.display.paradigm.matrix.display import MatrixDisplay
from bcipy.display.components.task_bar import CopyPhraseTaskBar


class MatrixCopyPhraseTask(RSVPCopyPhraseTask):
    """Matrix Copy Phrase Task.

    Initializes and runs all needed code for executing a copy phrase task. A
        phrase is set in parameters and necessary objects (daq, display) are
        passed to this function.

    Parameters
    ----------
        win : object,
            display window to present visual stimuli.
        daq : object,
            data acquisition object initialized for the desired protocol
        parameters : dict,
            configuration details regarding the experiment. See parameters.json
        file_save : str,
            path location of where to save data from the session
        signal_model : loaded pickle file,
            trained signal model.
        language_model: object,
            trained language model.
        fake : boolean, optional
            boolean to indicate whether this is a fake session or not.
    Returns
    -------
        file_save : str,
            path location of where to save data from the session
    """
    TASK_NAME = 'Matrix Copy Phrase Task'
    MODE = 'Matrix'

    def init_display(self) -> Display:
        """Initialize the Matrix display"""
        return init_display(self.parameters, self.window,
                            self.experiment_clock, self.spelled_text)


def init_display(parameters, win, experiment_clock, starting_spelled_text):
    """Constructs a new Matrix display"""
    # preview_inquiry = PreviewInquiryProperties(
    #     preview_only=parameters['preview_only'],
    #     preview_inquiry_length=parameters['preview_inquiry_length'],
    #     preview_inquiry_key_input=parameters['preview_inquiry_key_input'],
    #     preview_inquiry_progress_method=parameters[
    #         'preview_inquiry_progress_method'],
    #     preview_inquiry_isi=parameters['preview_inquiry_isi'])

    info = InformationProperties(
        info_color=[parameters['info_color']],
        info_pos=[(parameters['info_pos_x'], parameters['info_pos_y'])],
        info_height=[parameters['info_height']],
        info_font=[parameters['font']],
        info_text=[parameters['info_text']],
    )

    stimuli = StimuliProperties(stim_font=parameters['font'],
                                stim_pos=(parameters['stim_pos_x'],
                                          parameters['stim_pos_y']),
                                stim_height=parameters['stim_height'],
                                is_txt_stim=parameters['is_txt_stim'],
                                prompt_time=parameters['time_prompt'])

    task_bar = CopyPhraseTaskBar(win,
                                 task_text=parameters['task_text'],
                                 spelled_text=starting_spelled_text,
                                 colors=[parameters['task_color']],
                                 font=parameters['font'],
                                 height=parameters['task_height'])

    return MatrixDisplay(
        win,
        experiment_clock,
        stimuli,
        task_bar,
        info,
        trigger_type=parameters['trigger_type'],
        should_prompt_target=False)



import sys
from bcipy.gui.main import BCIGui, app, AlertMessageType, AlertMessageResponse

from bcipy.config import DEFAULT_FIELD_PATH, FIELD_FILENAME
from bcipy.helpers.load import load_fields
from bcipy.helpers.save import save_field_data


class FieldRegistry(BCIGui):
    """Field Registry.

    User interface for creating new fields for use in BCInterface.py.
    """

    padding = 100
    btn_height = 50
    default_text = '...'
    txt_color = 'white'
    input_text_color = 'gray'
    input_color = 'white'
    alert_title = 'Field Registry Alert'
    alert_timeout = 10
    field_types = {
        'Text': 'str',
        'True/False or Yes/No': 'bool',
        'Whole Number': 'int',
        'Decimal Point Number': 'float'
    }

    def __init__(self, *args, **kwargs):
        super(FieldRegistry, self).__init__(*args, **kwargs)

        # Structure of an field:
        #   { name: { help_text: '', type: int} }
        self.update_field_data()

        # These are set in the build_inputs and represent text inputs from the user
        self.name_input = None
        self.helptext_input = None
        self.type_input = None

        self.name = None
        self.helptext = None
        self.type = None

    def build_text(self) -> None:
        """Build Text.

        Build all static text needed for the UI.
        Positions are relative to the height / width of the UI defined in start_app.
        """
        text_x = 25
        text_y = 70
        font_size = 18
        self.add_static_textbox(
            text='Create BciPy Field',
            position=[self.width / 2 - self.padding, 0],
            size=[300, 100],
            background_color=self.background_color,
            text_color=self.txt_color,
            font_size=22
        )
        self.add_static_textbox(
            text='Name',
            position=[text_x, text_y],
            size=[200, 50],
            background_color=self.background_color,
            text_color=self.txt_color,
            font_size=font_size)
        text_y += self.padding
        self.add_static_textbox(
            text='Help Text',
            position=[text_x, text_y],
            size=[300, 50],
            background_color=self.background_color,
            text_color=self.txt_color,
            font_size=font_size)
        text_y += self.padding
        self.add_static_textbox(
            text='Type',
            position=[text_x, text_y],
            size=[300, 50],
            background_color=self.background_color,
            text_color=self.txt_color,
            font_size=font_size)

    def build_inputs(self) -> None:
        """Build Inputs.

        Build all text entry inputs for the UI.
        """
        input_x = 50
        input_y = 120
        input_size = [280, 25]
        self.name_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=[self.default_text],
            editable=True,
            background_color=self.input_color,
            text_color=self.input_text_color)

        input_y += self.padding
        self.helptext_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=[self.default_text],
            editable=True,
            background_color=self.input_color,
            text_color=self.input_text_color)

        input_y += self.padding
        self.type_input = self.add_combobox(
            position=[input_x, input_y],
            size=input_size,
            items=[],
            editable=False,
            background_color=self.input_color,
            text_color=self.input_text_color)

        self.update_type_list()

    def build_buttons(self):
        """Build Buttons.

        Build all buttons necessary for the UI. Define their action on click using the named argument action.
        """
        btn_create_x = self.width - self.padding - 10
        btn_create_y = self.height - 100
        size = 150
        self.add_button(
            message='Create Field', position=[btn_create_x - (size / 2), btn_create_y],
            size=[size, self.btn_height],
            background_color='green',
            action=self.create_field,
            text_color='white')

    def create_field(self) -> None:
        """Create field.

        After inputing all required fields, verified by check_input, add it to the field list and save it.
        """
        if self.check_input():
            self.add_field()
            self.save_fields()
            self.throw_alert_message(
                title=self.alert_title,
                message='Field saved successfully! Please exit window or create another field!',
                message_type=AlertMessageType.INFO,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.alert_timeout
            )
            self.refresh_inputs()

    def refresh_inputs(self):
        """Refresh Inputs.

        Clear and resets all inputs and related data.
        """
        self.name_input.clear()
        self.name_input.addItem(FieldRegistry.default_text)
        self.update_field_data()
        self.helptext_input.clear()
        self.helptext_input.addItem(FieldRegistry.default_text)
        self.update_type_list()

    def update_field_data(self):
        """Update Field Data.

        Fetches the fields and extracts the registered names.
        """
        self.fields = load_fields()
        self.field_names = list(self.fields.keys())

    def update_type_list(self):
        """Update Type List.

        Update the supported type list, reseting to default_text.
        """
        self.type_input.clear()
        self.type_input.addItem(FieldRegistry.default_text)
        self.type_input.addItems(list(self.field_types.keys()))

    def add_field(self) -> None:
        """Add field:

        Add a new field to the dict of fields. It follows the format:
             { name: { help_text: '', type: int} }
        """
        self.fields[self.name] = {
            'help_text': self.helptext,
            'type': self.field_types[self.type]
        }

    def save_fields(self) -> None:
        """Save field.

        Save the fields registered to the correct path as pulled from system_utils.
        """
        # add fields to the field
        save_field_data(self.fields, DEFAULT_FIELD_PATH, FIELD_FILENAME)

    def build_assets(self) -> None:
        """Build Assets.

        Define the assets to build in the UI.
        """
        self.build_inputs()
        self.build_text()
        self.build_buttons()

    def check_input(self) -> bool:
        """Check Input.

        Checks to make sure user has input all required fields. Currently, only name and summary are required.
        """
        self.name = self.name_input.currentText()
        self.helptext = self.helptext_input.currentText()
        self.type = self.type_input.currentText()

        try:
            if self.name == FieldRegistry.default_text or \
                    self.name == '':
                self.throw_alert_message(
                    title=self.alert_title,
                    message='Please add a field name!',
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
            if self.name in self.field_names:
                self.throw_alert_message(
                    title=self.alert_title,
                    message=(
                        'Field name already registered. \n'
                        'Please use a unique field name! \n'
                        f'Registed names: {self.field_names}'
                    ),
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
            if self.helptext == FieldRegistry.default_text or \
                    self.helptext == '':
                self.throw_alert_message(
                    title=self.alert_title,
                    message='Please add help text to field. \nThis will present when collecting experiment data!',
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
            if self.type == FieldRegistry.default_text:
                self.throw_alert_message(
                    title=self.alert_title,
                    message='Please select a valid field type!',
                    message_type=AlertMessageType.WARN,
                    message_response=AlertMessageResponse.OTE,
                    message_timeout=self.alert_timeout)
                return False
        except Exception as e:
            self.throw_alert_message(
                title=self.alert_title,
                message=f'Error, {e}',
                message_type=AlertMessageType.CRIT,
                message_response=AlertMessageResponse.OTE,
                message_timeout=self.alert_timeout)
            return False
        return True


def start_app() -> None:
    """Start Field Registry."""
    bcipy_gui = app(sys.argv)
    ex = FieldRegistry(
        title='Field Registry',
        height=500,
        width=550,
        background_color='black')

    ex.show_gui()

    sys.exit(bcipy_gui.exec_())


if __name__ == '__main__':
    start_app()



from typing import Optional, Tuple

import numpy as np
from bcipy.signal.process.filter import Notch, Bandpass


class Composition:
    """Applies a sequence of transformations"""

    def __init__(self, *transforms):
        self.transforms = transforms

    def __call__(self, data: np.ndarray, fs: Optional[int] = None) -> Tuple[np.ndarray, int]:
        for transform in self.transforms:
            data, fs = transform(data, fs)
        return data, fs


class Downsample:
    """Downsampling by an integer factor"""

    def __init__(self, factor: int = 2):
        self.factor = factor

    def __call__(self, data: np.ndarray, fs: Optional[int] = None) -> Tuple[np.ndarray, int]:
        if fs:
            return data[:, :: self.factor], fs // self.factor
        else:
            return data[:, :: self.factor], None


def get_default_transform(
    sample_rate_hz: int,
    notch_freq_hz: int,
    bandpass_low: int,
    bandpass_high: int,
    bandpass_order: int,
    downsample_factor: int,
    notch_quality_factor: int = 30,
) -> Composition:
    return Composition(
        Notch(sample_rate_hz, notch_freq_hz, notch_quality_factor),
        Bandpass(bandpass_low, bandpass_high, sample_rate_hz, bandpass_order),
        Downsample(downsample_factor),
    )



"""Functionality for reading and writing raw signal data."""
import csv
from typing import List, Optional, TextIO, Tuple

import numpy as np
import pandas as pd

from bcipy.config import DEFAULT_ENCODING
from bcipy.helpers.exceptions import BciPyCoreException
from bcipy.signal.generator.generator import gen_random_data
from bcipy.signal.process import Composition

TIMESTAMP_COLUMN = 'timestamp'


class RawData:
    """Represents the raw data format used by BciPy. Used primarily for loading
    a raw data file into memory."""

    def __init__(self,
                 daq_type: Optional[str] = None,
                 sample_rate: Optional[int] = None,
                 columns: Optional[List[str]] = None,
                 column_types: Optional[List[str]] = None):
        self.daq_type = daq_type
        self.sample_rate = sample_rate
        self.columns = columns or []
        # accept a custom column type definition or default to all eeg type
        self.column_types = column_types
        self._rows = []
        self._dataframe = None

    @classmethod
    def load(cls, filename: str):
        """Constructs a RawData object by deserializing the given file.
        All data will be read into memory. If you want to lazily read data one
        record at a time, use a RawDataReader.

        Parameters
        ----------
        - filename : path to the csv file to read
        """
        return load(filename)

    @property
    def rows(self) -> List[List]:
        """Returns the data rows"""
        return self._rows

    @rows.setter
    def rows(self, value):
        self._rows = value
        self._dataframe = None

    @property
    def channels(self) -> List[str]:
        """Compute the list of channels. Channels are the numeric columns
        excluding the timestamp column."""

        # Start data slice at 1 to remove the timestamp column.
        return list(self.numeric_data.columns[1:])

    @property
    def numeric_data(self) -> pd.DataFrame:
        """Data for columns with numeric data. This is usually comprised of the
        timestamp column and device channels, excluding string triggers."""
        return self.dataframe.select_dtypes(exclude=['object'])

    @property
    def channel_data(self):
        """Data for columns with numeric data, excluding the timestamp column."""
        numeric_data = self.numeric_data

        numeric_vals = numeric_data.values
        numeric_column_count = numeric_vals.shape[1]
        # Start data slice at 1 to remove the timestamp column.
        return numeric_vals[:, 1:numeric_column_count].transpose()

    def by_channel(self, transform: Optional[Composition] = None) -> np.ndarray:
        """Data organized by channel.

        Optionally, it can apply a BciPy Composition to the data before returning using the transform arg.
        This will apply n tranformations to the data before returning. For an example Composition with
        EEG preprocessing, see bcipy.signal.get_default_transform().

        Returns
        ----------
        data: C x N numpy array with samples where C is the number of channels and N
        is number of time samples
        fs: resulting sample rate if any transformations applied"""

        data = self.channel_data
        fs = self.sample_rate

        if transform:
            data, fs = self.apply_transform(data, transform)

        return data, fs

    def by_channel_map(
            self,
            channel_map: List[int],
            transform: Optional[Composition] = None) -> Tuple[np.ndarray, List[str], List[str]]:
        """By Channel Map.

        Returns channels with columns removed if index in list (channel_map) is zero. The channel map must align
        with the numeric channels read in as self.channels. We assume most trigger or other string columns are
        removed, however some are numeric trigger columns from devices that will require filtering before returning
        data. Other cases could be dropping bad channels before running further analyses.

        Optionally, it can apply a BciPy Composition to the data before returning using the transform arg.
        This will apply n tranformations to the data before returning. For an example Composition with
        EEG preprocessing, see bcipy.signal.get_default_transform().
        """
        data, fs = self.by_channel(transform)
        channels_to_remove = [idx for idx, value in enumerate(channel_map) if value == 0]
        data = np.delete(data, channels_to_remove, axis=0)
        channels = np.delete(self.channels, channels_to_remove, axis=0).tolist()

        return data, channels, fs

    def apply_transform(self, data: np.ndarray, transform: Composition) -> Tuple[np.ndarray, float]:
        """Apply Transform.

        Using data provided as an np.ndarray, call the Composition with self.sample_rate to apply
            transformations to the data. This will return the transformed data and resulting sample rate.
        """
        return transform(data, self.sample_rate)

    @property
    def dataframe(self) -> pd.DataFrame:
        """Returns a dataframe of the row data."""
        if self._dataframe is None:
            self._dataframe = pd.DataFrame(data=self.rows,
                                           columns=self.columns)
        return self._dataframe

    @property
    def total_seconds(self) -> float:
        """Total recorded seconds, defined as the diff between the first and
        last timestamp."""
        frame = self.dataframe
        col = 'lsl_timestamp'
        return frame.iloc[-1][col] - frame.iloc[0][col]

    @property
    def total_samples(self) -> int:
        """Number of samples recorded."""
        return int(self.dataframe.iloc[-1]['timestamp'])

    def query(self,
              start: Optional[float] = None,
              stop: Optional[float] = None,
              column: str = 'lsl_timestamp') -> pd.DataFrame:
        """Query for a subset of data.

        Pararameters
        ------------
            start - find records greater than or equal to this value
            stop - find records less than or equal to this value
            column - column to compare to the given start and stop values

        Returns
        -------
        Dataframe for the given slice of data
        """
        dataframe = self.dataframe
        start = start or dataframe.iloc[0][column]
        stop = stop or dataframe.iloc[-1][column]
        mask = (dataframe[column] >= start) & (dataframe[column] <= stop)
        return dataframe[mask]

    def append(self, row: List):
        """Append the given row of data.

        Parameters
        ----------
        - row : row of data
        """
        assert len(row) == len(self.columns), "Wrong number of columns"
        self._rows.append(row)
        self._dataframe = None


def maybe_float(val):
    """Attempt to convert the given value to float. If conversion fails return
    as is."""
    try:
        return float(val)
    except ValueError:
        return val


class RawDataReader:
    """Lazily reads raw data from a file. Intended to be used as a ContextManager
    using Python's `with` keyword.

    Example usage:

    ```
    with RawDataReader(path) as reader:
        print(f"Data from ${reader.daq_type}")
        print(reader.columns)
        for row in reader:
            print(row)
    ```

    Parameters
    ----------
    - file_path : path to the csv file
    - convert_data : if True attempts to convert data values to floats;
    default is False
    """

    def __init__(self, file_path: str, convert_data: bool = False):
        self.file_path = file_path
        self._file_obj = None
        self.convert_data = convert_data

    def __enter__(self):
        self._file_obj = open(self.file_path, mode="r", encoding=DEFAULT_ENCODING)
        self.daq_type, self.sample_rate = read_metadata(self._file_obj)
        self._reader = csv.reader(self._file_obj)
        self.columns = next(self._reader)
        return self

    def __exit__(self, *args, **kwargs):
        """Exit the context manager. Close resources"""
        self._file_obj.close()

    def __iter__(self):
        return self

    def __next__(self):
        assert self._reader, "Reader must be initialized"
        row = next(self._reader)
        if self.convert_data:
            return list(map(maybe_float, row))
        return row


class RawDataWriter:
    """Writes a raw data file one row at a time without storing the records
    in memory. Intended to be used as a ContextManager using Python's `with`
    keyword.

    Example usage:

    ```
    with RawDataWriter(path, daq_type, sample_rate, columns) as writer:
        for row in data:
            writer.writerow(row)
    ```

    Parameters
    ----------
    - file_path : path to the csv file
    - daq_type : name of device
    - sample_rate : sample frequency in Hz
    - columns : list of column names
    """

    def __init__(self, file_path: str, daq_type: str, sample_rate: float,
                 columns: List[str]):
        self.file_path = file_path
        self.daq_type = daq_type
        self.sample_rate = sample_rate
        self.columns = columns
        self._file_obj = None

    def __enter__(self):
        """Enter the context manager. Initializes the underlying data file."""
        self._file_obj = open(self.file_path,
                              mode='w',
                              encoding=DEFAULT_ENCODING,
                              newline='')
        # write metadata
        self._file_obj.write(f"daq_type,{self.daq_type}\n")
        self._file_obj.write(f"sample_rate,{self.sample_rate}\n")

        # If flush is missing the previous content may be appended at the end.
        self._file_obj.flush()

        self._csv_writer = csv.writer(self._file_obj, delimiter=',')
        self._csv_writer.writerow(self.columns)

        return self

    def __exit__(self, *args, **kwargs):
        """Exit the context manager. Close resources"""
        self._file_obj.close()

    def writerow(self, row: List) -> None:
        assert self._csv_writer, "Writer must be initialized"
        assert len(row) == len(self.columns), "Wrong number of columns"
        self._csv_writer.writerow(row)

    def writerows(self, rows: List[List]) -> None:
        for row in rows:
            self.writerow(row)


def load(filename: str) -> RawData:
    """Reads the file at the given path and initializes a RawData object.
    All data will be read into memory. If you want to lazily read data one
    record at a time, use a RawDataReader.

    Parameters
    ----------
    - filename : path to the csv file to read
    """
    # Loading all data from a csv is faster using pandas than using the
    # RawDataReader.
    try:
        with open(filename, mode='r', encoding=DEFAULT_ENCODING) as file_obj:
            daq_type, sample_rate = read_metadata(file_obj)
            dataframe = pd.read_csv(file_obj)
            data = RawData(daq_type, sample_rate, list(dataframe.columns))
            data.rows = dataframe.values.tolist()
            return data
    except FileNotFoundError:
        raise BciPyCoreException(f"\nError loading BciPy RawData. Valid data not found at: {filename}")


def read_metadata(file_obj: TextIO) -> Tuple[str, float]:
    """Reads the metadata from an open raw data file and retuns the result as
    a tuple. Increments the reader.

    Parameters
    ----------
    - file_obj : open TextIO object

    Returns
    -------
    tuple of daq_type, sample_rate
    """
    daq_type = next(file_obj).strip().split(',')[1]
    sample_rate = float(next(file_obj).strip().split(",")[1])
    return daq_type, sample_rate


def write(data: RawData, filename: str):
    """Write the given raw data file.

    Parameters
    ----------
    - data : raw data object to write
    - filename : path to destination file.
    """
    with RawDataWriter(filename, data.daq_type, data.sample_rate,
                       data.columns) as writer:
        for row in data.rows:
            writer.writerow(row)


def settings(filename: str) -> Tuple[str, float, List[str]]:
    """Read the daq settings from the given data file

    Parameters
    ----------
    - filename : path to the raw data file (csv)

    Returns
    -------
    tuple of (acquisition type, sample_rate, columns)
    """

    with RawDataReader(filename) as reader:
        return reader.daq_type, reader.sample_rate, reader.columns


def sample_data(rows: int = 1000,
                ch_names: List[str] = ['ch1', 'ch2', 'ch3'],
                daq_type: str = 'SampleDevice',
                sample_rate: float = 256.0,
                triggers: List[Tuple[float, str]] = []) -> RawData:
    """Creates sample data to be written as a raw_data.csv file. The resulting data has
    a column for the timestamp, one for each channel, and a TRG column.

    - rows : number of sample rows to generate
    - ch_names : list of channel names
    - daq_type : metadata for the device name
    - sample_rate : device sample rate in hz
    - triggers : List of (timestamp, trigger_value) tuples to be inserted
    in the data.
    """
    channels = [name for name in ch_names if name != 'TRG']
    columns = [TIMESTAMP_COLUMN] + channels + ['TRG']
    trg_times = dict(triggers)

    data = RawData(daq_type, sample_rate, columns)

    for i in range(rows):
        timestamp = i + 1
        channel_data = gen_random_data(low=-1000,
                                       high=1000,
                                       channel_count=len(channels))
        trg = trg_times.get(timestamp, '0.0')
        data.append([timestamp] + channel_data + [trg])

    return data
